(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

(*The soundness of all the examples is shown with the use of the pierogi proof rules. 
The soundness proof consists of 4 subproofs
 1)Initialisation:    Ensures that the initial assertion of each thread holds at the beginning.
 2)Local Correctness: Establishes annotation validity for each thread.
 3)Stability:         Ensures that each (local)thread annotation is interference-free under the execution of other threads.
 4)Persistence:       Ensures that the crash invariant holds at every program point.*)

(*Buffering with flushopts litmus test*)
(*Program outline
  x := 1       |   y := 1 
  flush\<^sub>o y     |   flush\<^sub>o x
  sfence       |   sfence
  z := 1       |   w := 1 
Crash invariant:  [w]\<^sub>P = {1} \<and>  [z]\<^sub>P  = {1} \<longrightarrow> ([x]\<^sub>P \<subseteq> {0,1} \<or>  [y]\<^sub>P \<subseteq> {0,1}) *)
 


theory Asynchronous_Flush_Buffering
imports "../pierogi/Proof_Rules" 
begin

datatype PC =
  L1
| L2
| L3
| L4
| L5

consts t1 ::  TId 
consts t2 ::  TId
consts x :: Loc
consts y :: Loc
consts z :: Loc
consts w :: Loc

definition "thrdsvars \<equiv> t1 \<noteq> t2 \<and> 
                         x \<noteq> y  \<and>
                         z \<noteq> y \<and>
                         z \<noteq> x  \<and>
                         w \<noteq> x \<and>
                         w \<noteq> y \<and>
                         w \<noteq> z  " 

(*the state of the program*)
record mp_state =
  pc :: "TId \<Rightarrow> PC"
  state :: TState
  aux1 :: nat 
  aux2 :: nat 

(*auxiliary definition for updating the pc*)
definition 
  "update_pc t nv pcf \<equiv> pcf (t := nv)"

lemmas mp_sis [simp] = 
  update_pc_def 
  thrdsvars_def

(*program encoding*)
definition prog :: "TId \<Rightarrow>  mp_state \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog t ts ts' \<equiv> 
  if t = t1
   then
    (case (pc ts) t of 
      L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [x:=1]\<^sub>t (state ts') 
   
    | L2 \<Rightarrow> pc ts' = update_pc t L3 (pc ts) \<and>
     (state ts)  [flush\<^sub>o y]\<^sub>t  (state ts')

    | L3 \<Rightarrow> pc ts' = update_pc t L4 (pc ts) \<and>
     (state ts)  [sfence]\<^sub>t  (state ts')

    | L4 \<Rightarrow>  pc ts' = update_pc t L5 (pc ts) \<and>
       (state ts) [z:=1]\<^sub>t (state ts') 
     
    | _ \<Rightarrow> ts = ts')
  else 
    if t = t2
    then 
  (case (pc ts) t of 

     L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [y:=1]\<^sub>t (state ts')
 
    | L2 \<Rightarrow>  pc ts' = update_pc t L3 (pc ts) \<and>
               (state ts)  [flush\<^sub>o x]\<^sub>t  (state ts') 

    | L3 \<Rightarrow> pc ts' = update_pc t L4 (pc ts) \<and>
     (state ts)  [sfence]\<^sub>t  (state ts')

    | L4 \<Rightarrow> pc ts' = update_pc t L5 (pc ts) \<and>
       (state ts) [w:=1]\<^sub>t (state ts') 

    |_ \<Rightarrow> ts = ts'

  )
    else ts = ts'"



(*program invariant*)
definition prog_inv :: "TId \<Rightarrow> PC \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog_inv t p ts \<equiv>
 if t = t1
   then 

   (case p of
          L1 \<Rightarrow>(  [z]\<^sub>P (state ts) =  {0}) 
              \<and> ( [y]\<^sub>t (state ts) \<subseteq> {0,1} )
 
        | L2 \<Rightarrow> (  [z]\<^sub>P (state ts) =  {0}) 
              \<and> ( [y]\<^sub>t (state ts) \<subseteq> {0,1} )

        | L3 \<Rightarrow>  (  [z]\<^sub>P (state ts) =  {0}) 
              \<and> (  [y]\<^sup>A\<^sub>t (state ts) \<subseteq> {0,1} )

        | L4 \<Rightarrow>   (  [z]\<^sub>P (state ts) =  {0}) 
              \<and> (  [y]\<^sub>P (state ts) \<subseteq> {0,1} )

       | L5 \<Rightarrow>   (  [z]\<^sub>P (state ts) =  {0}) 
             \<or> (  [y]\<^sub>P (state ts) \<subseteq> {0,1} )


      )
  else
 if t = t2 then
   (case p of
          L1 \<Rightarrow>(  [w]\<^sub>P (state ts) =  {0}) 
              \<and> ( [x]\<^sub>t (state ts) \<subseteq> {0,1} )
 
        | L2 \<Rightarrow> (  [w]\<^sub>P (state ts) =  {0}) 
              \<and> ( [x]\<^sub>t (state ts) \<subseteq> {0,1} )

        | L3 \<Rightarrow>  (  [w]\<^sub>P (state ts) =  {0}) 
              \<and> (  [x]\<^sup>A\<^sub>t (state ts) \<subseteq> {0,1} )

        | L4 \<Rightarrow>   (  [w]\<^sub>P (state ts) =  {0}) 
              \<and> (  [x]\<^sub>P (state ts) \<subseteq> {0,1} )

        | L5 \<Rightarrow>   (  [w]\<^sub>P (state ts) =  {0}) 
              \<or> (  [x]\<^sub>P (state ts) \<subseteq> {0,1} )

     
      )     
  else False " 



(*Initial State*)
definition init  :: "mp_state  \<Rightarrow> bool"
  where
    "init s \<equiv> (pc s) t1 = L1 
      \<and> (pc s) t2 = L1
      \<and> aux1 s  = 0
      \<and> aux2 s = 0
      \<and> start (state s)"


(*Local Correctness of thread 1*)
lemma t1_local:
  assumes "thrdsvars"
      and "total_wfs  (state s)"
      and "prog_inv t1 ((pc s) t1) s" 
      and "prog t1 s s'"
    shows "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t1", simp_all)
     apply (metis st_opv_daddr_ni st_ov_ni)
  apply (smt (verit, ccfv_SIG) assms(2) flo_oav_ov_rel flo_opv_ni subset_trans total_wfs_def)
  apply (smt (verit, best) OPV_notempty_preserved assms(2) opv_oav_sf sf_opv_sub subset_singleton_iff subset_trans total_wfs_def)
    by (metis st_opv_daddr_ni)

(*Local Correctness of thread 2*)
lemma t2_local:
  assumes "thrdsvars"
      and "total_wfs (state s)"
      and "prog_inv t2 ((pc s) t2) s" 
      and "prog t2 s s'"
    shows "prog_inv t2 ((pc s') t2) s'"
   using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t2", simp_all)
     apply (metis st_opv_daddr_ni st_ov_ni)
  apply (smt (verit, ccfv_SIG) assms(2) flo_oav_ov_rel flo_opv_ni subset_trans total_wfs_def)
  apply (smt (verit, best) OPV_notempty_preserved assms(2) opv_oav_sf sf_opv_sub subset_singleton_iff subset_trans total_wfs_def)
    by (metis st_opv_daddr_ni)

(*Stability check for thread 1*)
lemma t1_global :
  assumes  "total_wfs  (state s)"
  and  "prog_inv t1 ((pc s) t1) s"
  and "prog_inv t2 ((pc s) t2) s"
  and "thrdsvars"
  and "prog t2 s s'"
shows  "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_inv_def prog_def total_wfs_def  )
  apply (cases "pc s t1"; cases "pc s t2"; simp )
                     apply (smt (verit, ccfv_SIG) Un_subset_iff st_opv_daddr_ni st_ov_dt_lc subset_insertI)
                    apply (metis flo_opv_ni flo_ov_ni)
                   apply (metis OPV_notempty_preserved assms(1) sf_opv_sub sf_ov_ni subset_singleton_iff total_wfs_def)
                  apply (metis st_opv_daddr_ni st_ov_ni)
                 apply (smt (verit, best) Un_subset_iff st_opv_daddr_ni st_ov_dt_lc subset_insertI)
                apply (metis flo_opv_ni flo_ov_ni)
               apply (metis OPV_notempty_preserved assms(1) sf_opv_sub sf_ov_ni subset_singleton_iff total_wfs_def)
              apply (metis st_opv_daddr_ni st_ov_ni)
             apply (smt (z3) Un_subset_iff st_oav_gen st_opv_daddr_ni subset_insertI)
            apply (intro conjI)
             apply (metis flo_opv_ni)
            apply (metis flo_oav_daddr_ni)
  apply (metis OPV_notempty_preserved assms(1) sf_oav_ni sf_opv_sub subset_singleton_iff total_wfs_def)
          apply (metis st_oav_daddr_ni st_opv_daddr_ni)      
         apply (smt (verit, ccfv_SIG) Un_least st_opv_daddr_ni st_opv_gen subset_insertI)
        apply (metis flo_opv_ni)
  apply (smt (verit, ccfv_SIG) OPV_notempty_preserved assms(1) sf_opv_sub subset_singleton_iff subset_trans total_wfs_def)
      apply (metis st_opv_daddr_ni)
     apply (smt (verit, best) Un_subset_iff st_opv_daddr_ni st_opv_gen subset_insertI)
    apply (metis flo_opv_ni)
   apply (smt (verit, ccfv_SIG) OPV_notempty_preserved assms(1) sf_opv_sub subset_singleton_iff subset_trans total_wfs_def)
  by (metis st_opv_daddr_ni)
 


(*Stability check for thread 2*)
lemma t2_global :
  assumes  "total_wfs  (state s)"
  and "prog_inv t2 ((pc s) t2) s"
  and "prog_inv t1 ((pc s) t1) s"
  and "thrdsvars"
  and "prog t1 s s'"
shows  "prog_inv t2 ((pc s') t2) s'"
  using assms 
  apply (simp add: prog_inv_def prog_def total_wfs_def  )
  apply (cases "pc s t2"; cases "pc s t1"; simp )
                     apply (smt (verit, ccfv_SIG) Un_subset_iff st_opv_daddr_ni st_ov_dt_lc subset_insertI)
                    apply (metis flo_opv_ni flo_ov_ni)
                   apply (metis OPV_notempty_preserved assms(1) sf_opv_sub sf_ov_ni subset_singleton_iff total_wfs_def)
                  apply (metis st_opv_daddr_ni st_ov_ni)
                 apply (smt (verit, best) Un_subset_iff st_opv_daddr_ni st_ov_dt_lc subset_insertI)
                apply (metis flo_opv_ni flo_ov_ni)
               apply (metis OPV_notempty_preserved assms(1) sf_opv_sub sf_ov_ni subset_singleton_iff total_wfs_def)
              apply (metis st_opv_daddr_ni st_ov_ni)
             apply (smt (z3) Un_subset_iff st_oav_gen st_opv_daddr_ni subset_insertI)
            apply (intro conjI)
             apply (metis flo_opv_ni)
            apply (metis flo_oav_daddr_ni)
  apply (metis OPV_notempty_preserved assms(1) sf_oav_ni sf_opv_sub subset_singleton_iff total_wfs_def)
          apply (metis st_oav_daddr_ni st_opv_daddr_ni)      
         apply (smt (verit, ccfv_SIG) Un_least st_opv_daddr_ni st_opv_gen subset_insertI)
        apply (metis flo_opv_ni)
  apply (smt (verit, ccfv_SIG) OPV_notempty_preserved assms(1) sf_opv_sub subset_singleton_iff subset_trans total_wfs_def)
      apply (metis st_opv_daddr_ni)
     apply (smt (verit, best) Un_subset_iff st_opv_daddr_ni st_opv_gen subset_insertI)
    apply (metis flo_opv_ni)
   apply (smt (verit, ccfv_SIG) OPV_notempty_preserved assms(1) sf_opv_sub subset_singleton_iff subset_trans total_wfs_def)
  by (metis st_opv_daddr_ni)


(*Initialization*)
theorem init_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "init s"
shows "prog_inv t ((pc s) t) s"
  using assms 
  apply (simp add: init_def start_def prog_inv_def mapval_def OTS_def st_OTSF )   
  using st_OPTS start_def by auto



theorem final_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "t' \<in> {t1, t2}"
  and "prog_inv t ((pc s) t) s"
  and "prog_inv t' ((pc s) t') s" 
  and "prog t' s s'"
shows "prog_inv t ((pc s') t) s'"
  using assms apply (simp del: thrdsvars_def) 
  using t1_global t1_local t2_global t2_local
  by blast


(*Persistence*)
lemma persistence: "thrdsvars \<Longrightarrow> prog_inv t2 ((pc s) t2)  ( s) \<Longrightarrow> 
  [w]\<^sub>P (state s) = {1 } \<and>  [z]\<^sub>P (state s) = {1} \<longrightarrow> ([x]\<^sub>P (state s) \<subseteq> {0,1 } \<or>  [y]\<^sub>P (state s) \<subseteq> {0,1 })"
   apply (simp add: prog_inv_def)
  by (case_tac "pc s t2", auto )

end




