(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

(*The soundness of all the examples is shown with the use of the pierogi proof rules. 
The soundness proof consists of 4 subproofs
 1)Initialisation:    Ensures that the initial assertion of each thread holds at the beginning.
 2)Local Correctness: Establishes annotation validity for each thread.
 3)Stability:         Ensures that each (local)thread annotation is interference-free under the execution of other threads.
 4)Persistence:       Ensures that the crash invariant holds at every program point.*)

(*Message Passing with flushopts litmus test*)
(*Program outline 
 data := 42  | r1  \<leftarrow> flag
 flag := 7   | if (r1 = 7) 
             |   flush\<^sub>o data
             |   sfence
             |   commit := 1
Crash invariant: [commit]\<^sub>P = {1}  \<longrightarrow> [data]\<^sub>P = {42}
*)

theory Asynchronous_Message_Passing
  imports   "../pierogi/Proof_Rules" 
begin

datatype PC =
  L1
| L2
| L3
| L4
| L5
|L6

consts t1 ::  TId 
consts t2 ::  TId
consts data :: Loc
consts flag :: Loc
consts commit :: Loc
consts r1 :: Reg
consts r2 :: Reg 

definition "thrdsvars \<equiv> t1 \<noteq> t2 \<and> 
                        data \<noteq> flag  \<and>
                        commit \<noteq> flag \<and>
                        commit \<noteq> data" 

(*the state of the program*)
record mp_state =
  pc :: "TId \<Rightarrow> PC"
  state :: TState


(*auxiliary definition for updating the pc*)
definition 
  "update_pc t nv pcf \<equiv> pcf (t := nv)"

lemmas mp_sis [simp] = 
  update_pc_def 
  thrdsvars_def

(*program encoding*)
definition prog :: "TId \<Rightarrow>  mp_state \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog t ts ts' \<equiv> 
  if t = t1
   then
    (case (pc ts) t of 
      L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
       (state ts) [data:=42]\<^sub>t (state ts') 
    | L2 \<Rightarrow> pc ts' = update_pc t L3 (pc ts) \<and>
              (state ts)  [flag:=7]\<^sub>t  (state ts')
    | _ \<Rightarrow> ts = ts')

    else 
    if t = t2
    then (case (pc ts) t of 
       L1 \<Rightarrow> pc ts' = update_pc t L2 (pc ts) \<and>
            (state ts) [r1   \<leftarrow> flag ]\<^sub>t (state ts') 
    |  L2 \<Rightarrow> if  (regs (state ts) t2  r1  =  7) then
                pc ts' = update_pc t L3 (pc ts) \<and>
                 state ts = state ts'
              
            else pc ts' = update_pc t L6 (pc ts) \<and>
                 state ts = state ts'
    |  L3 \<Rightarrow>  pc ts' = update_pc t L4 (pc ts)  \<and>   
            (state ts)    [flush\<^sub>o data ]\<^sub>t   (state ts')
    |  L4 \<Rightarrow>  pc ts' = update_pc t L5 (pc ts)  \<and>
       (state ts)  [sfence]\<^sub>t  (state ts')
    |  L5 \<Rightarrow>  pc ts' = update_pc t L6 (pc ts) \<and>
           (state ts)  [commit:=1]\<^sub>t  (state ts')
    |  _ \<Rightarrow> ts = ts')
    else ts = ts'"


(*program invariant*)
definition prog_inv :: "TId \<Rightarrow> PC \<Rightarrow> TState \<Rightarrow> bool " where
"prog_inv t p ts \<equiv>
 if t = t1
   then 

   (case p of
          L1 \<Rightarrow>  7 \<notin> [flag]\<^sub>t2  ts 
        | L2 \<Rightarrow>  [data]\<^sub>t  ts = { 42 } \<and>  7 \<notin> [flag]\<^sub>t2  ts
        | L3 \<Rightarrow> True
        | L4 \<Rightarrow> True
        | L5 \<Rightarrow> True
        | L6 \<Rightarrow> True
      )
  else
 if t = t2 then
      (case p of
          L1 \<Rightarrow>  (7 \<in> [flag]\<^sub>t ( ts) \<longrightarrow> \<langle>flag  7\<rangle>[data]\<^sub>t ( ts) = {42}) \<and> [flag]\<^sub>t ( ts) \<subseteq> { 0,7} \<and>  [commit]\<^sub>P ( ts) = {0}
        | L2 \<Rightarrow>  ( ( (regs ( ts) t  r1  = 7) \<longrightarrow> [data]\<^sub>t ( ts) = {42}) \<and>  [commit]\<^sub>P ( ts) = {0}  )
        | L3 \<Rightarrow>   ( ( [data]\<^sub>t ( ts) = {42}) \<and>  [commit]\<^sub>P ( ts) = {0}  )
        | L4 \<Rightarrow>   [data]\<^sup>A\<^sub>t ( ts) = {42} \<and>  [commit]\<^sub>P ( ts) = {0}
        | L5 \<Rightarrow>    [data]\<^sub>P (ts) = {42} 
        | L6 \<Rightarrow>   [data]\<^sub>P (ts) = {42} \<or>   [commit]\<^sub>P ( ts) = {0}  
      )       
  else False "

(*Initial State*)
definition init  :: "mp_state  \<Rightarrow> bool"
  where
    "init s \<equiv> (pc s) t1 = L1 
      \<and> (pc s) t2 = L1
    \<and> start (state s)"

(*Local Correctness of thread 1*)
lemma t1_local:
  assumes "thrdsvars"
      and "total_wfs  (state s)"
      and "prog_inv t1 ((pc s) t1) (state s)" 
      and "prog t1 s s'"
    shows "prog_inv t1 ((pc s') t1) (state s')"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t1", simp_all)
  by (metis st_ov_lc st_ov_ni)
 
(*Local Correctness of thread 2*)
lemma t2_local:
  assumes "thrdsvars"
      and "total_wfs (state s)"
      and "prog_inv t2 ((pc s) t2) (state s)" 
      and "prog t2 s s'"
    shows "prog_inv t2 ((pc s') t2) (state s')"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t2", simp_all)
  apply (rule conjI)
       apply (simp add: total_OTSF_def)
  apply (metis assms(2) ld_cobv_lc_s ld_ov_lc total_wfs_def)
     apply (smt (verit) empty_iff insert_iff ld_opv_ni ld_ov_lc subset_iff)
     apply (simp split:if_splits )
  using flo_oav_ov_rel_s flo_opv_ni apply force
  apply (metis assms(2) opv_oav_sf_s  total_wfs_def)
  by (metis st_opv_daddr_ni)
     

(*Stability check for thread 1*)
lemma t1_global :
  assumes  "total_wfs  (state s)"
  and  "prog_inv t1 ((pc s) t1)  (state s)"
  and "prog_inv t2 ((pc s) t2)  (state s)"
  and "thrdsvars"
  and "prog t2 s s'"
shows  "prog_inv t1 ((pc s') t1)  (state s')"
  using assms 
  apply (simp add: prog_inv_def prog_def total_wfs_def )
   apply (cases "pc s t1"; cases "pc s t2";   simp add: split if_split )
              apply (metis ld_nov_ni )
             apply (simp split:if_splits )
             apply (metis flo_ov_ni )
  using sf_ov_ni apply fastforce
  using st_ov_ni apply blast    
          apply (metis ld_nov_ni ld_ov_ni)
         apply (simp split:if_splits )
  apply (metis flo_ov_ni)
       apply (metis sf_ov_ni)
  apply (metis st_ov_ni)
     apply (simp split:if_splits )
    apply (simp split:if_splits )
   apply (simp split:if_splits )
  by (simp split:if_splits )

(*Stability check for thread 2*)
lemma t2_global :
  assumes  "total_wfs  (state s)"
  and "prog_inv t2 ((pc s) t2)  (state s)"
  and "prog_inv t1 ((pc s) t1)  (state s)"
  and "thrdsvars"
  and "prog t1 s s'"
shows  "prog_inv t2 ((pc s') t2)  (state s')"
  using assms 
   apply (simp add: prog_inv_def  prog_def total_wfs_def )
  apply (cases "pc s t2"; cases "pc s t1";   simp add: split )
            apply (metis st_opv_daddr_ni st_ov_ni)
  apply (smt (verit, best) Un_subset_iff assms(1) ld_cobv_lc_intro_s st_opv_daddr_ni st_ov_dt_lc subset_insertI total_OTSF_def total_wfs_def)
          apply (metis Un_absorb reg_same_st st_opv_daddr_ni st_ov_dt_lc)
         apply (metis reg_same_st st_opv_daddr_ni st_ov_ni)
        apply (metis Un_absorb st_opv_daddr_ni st_ov_dt_lc)
        apply (metis st_opv_daddr_ni st_ov_ni)
  apply (metis st_oav_gen st_opv_daddr_ni sup.idem)
      apply (metis st_oav_daddr_ni st_opv_daddr_ni)
  apply (metis Un_absorb  st_opv_gen)
    apply (metis st_opv_daddr_ni)
   apply (metis Un_absorb st_opv_daddr_ni st_opv_gen)
  by (metis st_opv_daddr_ni)

(*Initialization*)
theorem init_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "init s"
shows "prog_inv t ((pc s) t)  (state s)"
  using assms 
  apply (simp add: init_def start_def prog_inv_def )
  using st_OATS st_OPTS st_OTS st_OV_z start_def by force


theorem final_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "t' \<in> {t1, t2}"
  and "prog_inv t ((pc s) t) (state s)"
  and "prog_inv t' ((pc s) t') (state s)" 
  and "prog t' s s'"
shows "prog_inv t ((pc s') t) (state s')"
  using assms apply (simp del: thrdsvars_def) 
  using t1_global t1_local t2_global t2_local by blast

(*Persistence*)
lemma persistence:
  "thrdsvars \<Longrightarrow> prog_inv t2 ((pc s) t2)  (state s) \<Longrightarrow> 
  [commit]\<^sub>P (state s) = {1}  \<longrightarrow> [data]\<^sub>P (state s) = {42}"
   apply (simp add: prog_inv_def split:if_splits)
  by (case_tac "pc s t2", auto )


end


