(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

(*The soundness of all the examples is shown with the use of the pierogi proof rules. 
The soundness proof consists of 4 subproofs
 1)Initialisation:    Ensures that the initial assertion of each thread holds at the beginning.
 2)Local Correctness: Establishes annotation validity for each thread.
 3)Stability:         Ensures that each (local)thread annotation is interference-free under the execution of other threads.
 4)Persistence:       Ensures that the crash invariant holds at every program point.*)

(* CAS  with flushes litmus test*)
(*Program outline
  CAS lx 0 1 r1 |  CAS lx 0 2 r2
  if (r1 = 1)   |  if (r2 = 1)
     x:=1       |    f \<leftarrow> x
     y:=1       |    if (f = 1 \<and> r2 = 2) 
     flush x    |      z := 1
     flush y    |      flush z
     lx = 0     |   lx = 0
Crash invariant:  [z]\<^sub>P  = {1} \<longrightarrow> ([x]\<^sub>P = {1} \<and>  [y]\<^sub>P = {1}) *)



theory CAS_Example
imports  "../pierogi/Proof_Rules"
begin

datatype PC =
   L1
  |L2
  |L3
  |L4
  |L5
  |L6
  |L7
  |L8



consts t1 ::  TId 
consts t2 ::  TId
consts x :: Loc
consts y :: Loc
consts lx :: Loc
consts z :: Loc
consts r1 :: Reg
consts r2 :: Reg
consts f :: Reg



definition "thrdsvars \<equiv> t1 \<noteq> t2  \<and> 
                         x \<noteq> y  \<and>
                         z \<noteq> y \<and>
                         z \<noteq> x \<and>
                         lx \<noteq> x \<and>  
                         lx \<noteq> y \<and>
                         lx \<noteq> z \<and>  
                         r1 \<noteq> r2 \<and>                       
                         r2 \<noteq> f \<and>
                         r1 \<noteq> f
  " 


(*the state of the program*)
record mp_state =
  pc :: "TId \<Rightarrow> PC"
  state :: TState

(*auxiliary definition for updating the pc*)
definition 
  "update_pc t nv pcf \<equiv> pcf (t := nv)"

lemmas mp_sis [simp] = 
  update_pc_def 
  thrdsvars_def


(*program encoding*)
definition prog :: "TId \<Rightarrow>  mp_state \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog t ts ts' \<equiv> 
  if t = t1
   then
   (case (pc ts) t of 
      L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [CAS lx 0 1 r1 ]\<^sub>t(state ts')


      |L2 \<Rightarrow> if  (regs (state ts) t1  r1  =  1) then
                pc ts' = update_pc t L3 (pc ts) \<and>
                 state ts = state ts'            
            else pc ts' = update_pc t L8 (pc ts) \<and>
                 state ts = state ts' 
    |  L3 \<Rightarrow>  (state ts)  [x:=1]\<^sub>t  (state ts') \<and>
             pc ts' = update_pc t L4 (pc ts)

    |  L4 \<Rightarrow>  (state ts)  [y:=1]\<^sub>t  (state ts') \<and>
             pc ts' = update_pc t L5 (pc ts)

    |  L5 \<Rightarrow>  (state ts)  [flush x]\<^sub>t  (state ts') \<and>
             pc ts' = update_pc t L6 (pc ts)

    |  L6 \<Rightarrow>  (state ts)  [flush y]\<^sub>t  (state ts') \<and>      
             pc ts' = update_pc t L7 (pc ts)

   |  L7 \<Rightarrow>  (state ts)  [lx := 0]\<^sub>t  (state ts') \<and>
             pc ts' = update_pc t L8 (pc ts)
    | _ \<Rightarrow> ts = ts')
  else 
    if t = t2
    then 
  (case (pc ts) t of 

       L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
       (state ts) [CAS lx 0 2 r2 ]\<^sub>t(state ts')
 
     | L2 \<Rightarrow>  if  (regs (state ts) t2  r2  =  1) then
                pc ts' = update_pc t L3 (pc ts) \<and>
                 state ts = state ts'             
            else pc ts' = update_pc t L8 (pc ts) \<and>
                 state ts' = state ts

  |  L3 \<Rightarrow>  pc ts' = update_pc t L4 (pc ts) \<and>
        (state ts) [f   \<leftarrow> x ]\<^sub>t (state ts') 

  |  L4 \<Rightarrow>  if  (regs (state ts) t2  f  =  1 \<and> regs (state ts) t2  r2  =  1) then
                pc ts' = update_pc t L5 (pc ts) \<and>
                 state ts'= state ts            
            else  
             pc ts' = update_pc t L7 (pc ts)\<and>
            state ts'= state ts
       

  |  L5 \<Rightarrow>   (state ts)  [z := 1]\<^sub>t  (state ts') \<and>
             pc ts' = update_pc t L6 (pc ts)

  |  L6 \<Rightarrow>  (state ts)  [flush z]\<^sub>t  (state ts') \<and>
             pc ts' = update_pc t L7 (pc ts)

 |  L7 \<Rightarrow>  (state ts)  [lx := 0]\<^sub>t  (state ts') \<and>
             pc ts' = update_pc t L8 (pc ts)


    |_ \<Rightarrow> ts = ts'

  )
    else ts = ts'"

(*program invariant*)
definition prog_inv :: "TId \<Rightarrow> PC \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog_inv t p ts \<equiv>
 if t = t1
   then 
   (case p of
          L1 \<Rightarrow> ( \<lceil>lx:0\<rceil>  (state ts) \<and> regs  (state ts) t  r1  = 0 )  
             \<or> (  \<lceil>lx:2\<rceil>  (state ts) \<and>  regs  (state ts) t  r1  = 0 )
 
      | L2 \<Rightarrow> (  regs  (state ts) t  r1  = 1 \<and>  \<lceil>lx:1\<rceil>  (state ts)  ) \<or>
                (  regs  (state ts) t  r1  = 0  ) 

      | L3 \<Rightarrow> (   \<lceil>lx:1\<rceil>  (state ts)  )

      |  L4  \<Rightarrow> (   \<lceil>lx:1\<rceil>  (state ts) \<and> [x]\<^sub>t  (state ts) = {1} \<and> \<lceil>x, 1\<rceil>\<^sub>t1  (state ts) )

      |  L5  \<Rightarrow> (  \<lceil>lx:1\<rceil>  (state ts) \<and> [x]\<^sub>t  (state ts) = {1} \<and> [y]\<^sub>t  (state ts) = {1} \<and> \<lceil>x, 1\<rceil>\<^sub>t1  (state ts) )

      |  L6 \<Rightarrow> ( \<lceil>lx:1\<rceil>  (state ts) \<and> [x]\<^sub>P  (state ts) = {1} \<and> [y]\<^sub>t  (state ts) = {1}\<and> \<lceil>x, 1\<rceil>\<^sub>t1  (state ts)  ) 

     |  L7 \<Rightarrow> (  \<lceil>lx:1\<rceil>  (state ts) \<and> [x]\<^sub>P  (state ts) = {1} \<and> [y]\<^sub>P  (state ts) = {1}   \<and> \<lceil>x, 1\<rceil>\<^sub>t1  (state ts) ) 

     |  L8 \<Rightarrow> ( [x]\<^sub>P  (state ts) = {1} \<and> [y]\<^sub>P  (state ts) = {1}) \<or> ( regs  (state ts) t  r1  = 0 ) 
   

               
      )
  else
 if t = t2 then
    (case p of
          L1 \<Rightarrow> (  \<lceil>lx:0\<rceil>  (state ts) \<and> (1 \<notin>[x]\<^sub>t2 (state ts)) \<and> regs  (state ts) t  r2  = 0  \<and> [z]\<^sub>P  (state ts) = {0}  \<and> regs  (state ts) t  r2  = 0  \<and> regs  (state ts) t  f  = 0   ) \<or>
                (   \<lceil>lx:0\<rceil>  (state ts) \<and>  [x]\<^sub>P  (state ts) = {1} \<and> [y]\<^sub>P  (state ts) = {1}  \<and> regs  (state ts) t  r2  = 0  \<and> \<lceil>x, 1\<rceil>\<^sub>t1  (state ts)    \<and> [z]\<^sub>P  (state ts) = {0}  \<and> regs  (state ts) t  f  = 0 )
             \<or> ( \<lceil>lx:1\<rceil>  (state ts)  \<and> regs  (state ts) t  r2  = 0 \<and> [z]\<^sub>P  (state ts) = {0}  \<and> regs  (state ts) t  f  = 0  )
 

        | L2 \<Rightarrow> (   regs  (state ts) t  r2  = 1 \<and>  1 \<notin>[x]\<^sub>t2 (state ts)  \<and>  \<lceil>lx:2\<rceil>  (state ts)\<and> [z]\<^sub>P  (state ts) = {0}   ) \<or>
                (   regs  (state ts) t  r2  = 1 \<and>  [x]\<^sub>P  (state ts) = {1} \<and> [y]\<^sub>P  (state ts) = {1}  \<and>  \<lceil>lx:2\<rceil>  (state ts)\<and> [x]\<^sub>t2  (state ts)  = {1}\<and> [z]\<^sub>P  (state ts) = {0}    ) \<or>
                ( regs  (state ts) t  r2  = 0 \<and> [z]\<^sub>P  (state ts) = {0}  ) 

        |L3 \<Rightarrow> (    1 \<notin>[x]\<^sub>t2 (state ts)  \<and>  \<lceil>lx:2\<rceil>  (state ts) \<and> [z]\<^sub>P  (state ts) = {0}   ) \<or>
               (    [x]\<^sub>P  (state ts) = {1} \<and> [y]\<^sub>P  (state ts) = {1}  \<and>  \<lceil>lx:2\<rceil>  (state ts) \<and> [x]\<^sub>t2  (state ts)  = {1} \<and> [z]\<^sub>P  (state ts) = {0}  )

       |L4 \<Rightarrow> ( ( regs (state ts) t  f  =  1 \<longrightarrow> (    [x]\<^sub>P  (state ts) = {1} 
                                                  \<and> [y]\<^sub>P  (state ts) = {1}  \<and>  \<lceil>lx:2\<rceil> (state ts)  \<and> [x]\<^sub>t2  (state ts)  = {1}) )
             \<and> [z]\<^sub>P  (state ts) = {0} \<and>  \<lceil>lx:2\<rceil> (state ts))

      | L5 \<Rightarrow>   (   [x]\<^sub>P  (state ts) = {1} \<and> [y]\<^sub>P  (state ts) = {1}  \<and>  \<lceil>lx:2\<rceil>  (state ts)  )

      | L6 \<Rightarrow>   (    [x]\<^sub>P  (state ts) = {1} \<and> [y]\<^sub>P  (state ts) = {1}  \<and>  \<lceil>lx:2\<rceil>  (state ts) \<and> [z]\<^sub>t  (state ts) = {1}    )

     | L7 \<Rightarrow>   ((   [x]\<^sub>P  (state ts) = {1} \<and> [y]\<^sub>P  (state ts) = {1}   \<and> [z]\<^sub>P  (state ts) = {1 } \<and>  \<lceil>lx:2\<rceil>  (state ts) ) \<or>   (  [z]\<^sub>P  (state ts) = {0} \<and>  \<lceil>lx:2\<rceil>  (state ts) ) )


   | L8 \<Rightarrow>    ( ( [x]\<^sub>P  (state ts) = {1} \<and> [y]\<^sub>P  (state ts) = {1}   \<and> [z]\<^sub>P  (state ts) = {1} )  \<or>  [z]\<^sub>P  (state ts) = {0})


    )
  else False "  

(*Initial State*)
definition init  :: "mp_state  \<Rightarrow> bool"
  where
    "init s \<equiv> (pc s) t1 = L1 
      \<and> (pc s) t2 = L1
      \<and> start (state s)"


(*Local Correctness of thread 1*)
lemma t1_local:
  assumes "thrdsvars"
      and "total_wfs  (state s)"
      and "prog_inv t1 ((pc s) t1) s" 
      and "prog t1 s s'"
    shows "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t1";simp) 
  apply (metis (mono_tags, lifting) One_nat_def cas_lc compval_def)
       apply (simp split:if_splits )
      apply (metis nat.distinct(1)  st_last_entry_init_message_ni st_last_entry_val_ni st_ots_lm_lc st_ov_lc vbounded_preserved)
  apply (intro conjI)
  apply (metis not_less_zero st_last_entry_init_message_ni zero_less_Suc)
  apply (smt (verit, ccfv_threshold) compval_def st_lv_ni st_ots_lm_ni st_ov_lc st_ov_ni)
    apply (intro conjI)
   apply (metis fl_last_entry_init_message_ni old.nat.distinct(2))
    apply (intro conjI impI)   
        apply (metis fl_last_entry_val_ni nat.distinct(1))
       apply (meson assms(2) fl_opv_ov_rel_s total_wfs_def)
      apply (metis fl_ov_ni)
     apply (metis fl_ov_ni)
    apply (meson fl_ots_lm_ni)
 apply (intro conjI)
     apply (metis fl_last_entry_init_message_ni old.nat.distinct(2))
   apply (intro conjI impI)
       apply (metis fl_last_entry_val_ni nat.distinct(1))
      apply (metis fl_opv_daddr_ni)
     apply (meson assms(2) fl_opv_ov_rel_s total_wfs_def)
    apply (metis fl_ov_ni)
   apply (meson fl_ots_lm_ni)
  by (metis st_opv_daddr_ni)



lemmas [simp del] = compval_def


(*Local Correctness of thread 2*)
lemma t2_local:
  assumes "thrdsvars"
      and "total_wfs (state s)"
      and "prog_inv t2 ((pc s) t2) s" 
      and "prog t2 s s'"
    shows "prog_inv t2 ((pc s') t2) s'"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def thrdsvars_def)
  apply (simp split:if_splits  )
    apply (cases "pc s t2";simp)
   apply (elim disjE)
        apply (metis Load_Rules.st_lv_lc ld_opv_ni ld_ov_lc)
  using Load_Rules.st_lv_lc assms(2) ld_opv_ni ld_ots_s total_wfs_def apply force
   apply (metis (no_types, lifting) compval_def nat.discI numeral_2_eq_2 st_last_entry_init_message_ni st_last_entry_val_ni st_opv_daddr_ni st_ov_lc)
  apply (subgoal_tac " \<lceil>lx:2\<rceil> state s' \<and> [x]\<^sub>Pstate s' = {Suc 0} \<and>
    [y]\<^sub>Pstate s' = {Suc 0} \<and>
    [z]\<^sub>Pstate s' = {Suc 0}")
  prefer 2
      apply (rule conjI)
       apply (metis fl_lv_lc)
  using assms(2) fl_opv_daddr_ni fl_opv_ov_rel_s total_wfs_def apply force
  apply meson
  using st_opv_daddr_ni apply force
   apply (cases "pc s t2";simp)
  apply (smt (verit, ccfv_SIG) Load_Rules.st_lv_lc assms(2) ld_opv_ni ld_ots_s ld_ov_lc total_wfs_def)
  apply (metis (no_types, lifting) compval_def st_last_entry_init_message_ni st_last_entry_val_ni st_opv_daddr_ni st_ov_lc zero_neq_numeral)
  apply (subgoal_tac " \<lceil>lx:2\<rceil> state s' \<and> [x]\<^sub>Pstate s' = {Suc 0} \<and>
    [y]\<^sub>Pstate s' = {Suc 0} \<and>
    [z]\<^sub>Pstate s' = {Suc 0}")
  prefer 2
      apply (rule conjI)
       apply (metis fl_lv_lc)
  using assms(2) fl_opv_daddr_ni fl_opv_ov_rel_s total_wfs_def apply force
    apply meson
  using st_opv_daddr_ni apply force
  apply (cases "pc s t2";simp)
  apply (smt (verit) assms(2) cas_lc cas_lv_daddr_lc cas_nlv_lc cas_nov_daddr_ni cas_opv_ni numeral_1_eq_Suc_0 numerals(1) total_wfs_def)
     apply (smt (verit, ccfv_SIG) Load_Rules.st_lv_lc assms(2) ld_opv_ni ld_ots_s ld_ov_lc total_wfs_def)
  apply (metis (no_types, lifting) compval_def st_last_entry_init_message_ni st_last_entry_val_ni st_opv_daddr_ni st_ov_lc zero_neq_numeral)
  using assms(2) fl_lv_lc fl_opv_daddr_ni fl_opv_ov_rel_s total_wfs_def apply force
  using st_opv_daddr_ni by force

(*Stability check for thread 1*)
lemma t1_global :
  assumes  "total_wfs  (state s)"
  and  "prog_inv t1 ((pc s) t1) s"
  and "prog_inv t2 ((pc s) t2) s"
  and "thrdsvars"
  and "prog t2 s s'"
shows  "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_inv_def prog_def total_wfs_def  )
 apply (simp split:if_splits )
    apply (cases "pc s t1"; cases "pc s t2";simp)
  apply (metis Load_Rules.st_lv_lc reg_same_ld_dt)
                  apply (metis compval_def reg_same_st st_last_entry_val_ni zero_neq_numeral)
                 apply (metis fl_lv_lc reg_same_fl)
                apply (metis Store_Rules.st_lv_lc reg_same_st)
               apply (metis (mono_tags, lifting) n_not_Suc_n numeral_2_eq_2 reg_same_ld_dt)
              apply (metis reg_same_st)
             apply (metis reg_same_fl)
            apply (metis n_not_Suc_n numeral_2_eq_2 reg_same_st)
           apply metis
          apply metis
         apply metis
        apply metis
       apply (metis (mono_tags, lifting) ld_opv_ni reg_same_ld_dt)
      apply (metis st_opv_daddr_ni)
     apply (metis fl_opv_daddr_ni)
    apply (metis reg_same_st st_opv_daddr_ni)
   apply (cases "pc s t1"; cases "pc s t2";simp) 
  apply (metis Load_Rules.st_lv_lc reg_same_ld_dt)
                 apply (metis compval_def reg_same_st st_last_entry_val_ni zero_neq_numeral)
                apply (metis fl_lv_lc reg_same_fl)
               apply (metis Store_Rules.st_lv_lc reg_same_st) 
  apply (metis n_not_Suc_n numeral_2_eq_2 reg_same_ld_dt)
             apply (metis reg_same_st)
            apply (metis reg_same_fl)
           apply (metis n_not_Suc_n numeral_2_eq_2 reg_same_st)
          apply metis
         apply metis
        apply metis
       apply metis
      apply (metis (mono_tags, lifting) ld_opv_ni reg_same_ld_dt)
     apply (metis st_opv_daddr_ni)
    apply (metis fl_opv_daddr_ni)
  apply (metis reg_same_st st_opv_daddr_ni) 
  apply (cases "pc s t1"; cases "pc s t2";simp)
                      apply (smt (verit, best) cas_possible_lv_lc reg_same_CAS_dt)
  
                      apply (metis Load_Rules.st_lv_lc reg_same_ld_dt)  
                      apply (metis compval_def reg_same_st st_last_entry_val_ni zero_neq_numeral)
                      apply (metis fl_lv_lc reg_same_fl)
                      apply (metis Store_Rules.st_lv_lc reg_same_st)
                      apply (smt (verit, best) cas_fail_lv_ni reg_same_CAS_dt)
                      apply (metis (mono_tags, lifting) n_not_Suc_n numeral_2_eq_2 reg_same_ld_dt)
                      apply (metis reg_same_st)
                     apply (metis reg_same_fl)
                    apply (metis n_not_Suc_n numeral_2_eq_2 reg_same_st)
                   apply (metis cas_fail_lv_ni numeral_1_eq_Suc_0 numeral_eq_one_iff zero_neq_one)
  apply (smt (verit) assms(1) cas_fail_lv_ni cas_ots_lm_ni cas_ov_daddr_ni_s nat.discI total_wfs_def)
                 apply metis
                apply metis
  apply (smt (verit, best) assms(1) cas_fail_lv_ni cas_ots_lm_ni cas_ov_daddr_ni_s numeral_1_eq_Suc_0 numerals(1) total_wfs_def zero_neq_one)
  apply metis
             apply metis
  apply (smt (verit) assms(1) cas_fail_lv_ni cas_opv_ni cas_ots_lm_ni cas_ov_daddr_ni_s nat.discI total_wfs_def)
           apply metis
          apply metis
  apply (smt (verit) assms(1) cas_fail_lv_ni cas_opv_ni cas_ots_lm_ni cas_ov_daddr_ni_s nat.discI total_wfs_def)
  apply metis
       apply metis 
  apply (metis cas_opv_ni reg_same_CAS_dt)
     apply (metis (mono_tags, lifting) ld_opv_ni reg_same_ld_dt)
    apply (metis st_opv_daddr_ni)
   apply (metis fl_opv_daddr_ni)
  by (metis reg_same_st st_opv_daddr_ni)


(*Stability check for thread 2*)
lemma t2_global :
  assumes  "total_wfs  (state s)"
  and "prog_inv t2 ((pc s) t2) s"
  and "prog_inv t1 ((pc s) t1) s"
  and "thrdsvars"
  and "prog t1 s s'"
shows  "prog_inv t2 ((pc s') t2) s'"
  using assms 
  apply (simp add: prog_inv_def  prog_def total_wfs_def )
  apply (simp split:if_splits )
   apply (cases "(pc s) t2"; cases "pc s t1"; simp add: prog_inv_def  prog_def)
                 apply (metis reg_same_st st_lv_ni st_opv_daddr_ni)
                apply (metis reg_same_st st_lv_ni st_opv_daddr_ni)
               apply (metis fl_lv_lc fl_opv_daddr_ni reg_same_fl)
              apply (metis fl_lv_lc fl_opv_daddr_ni reg_same_fl)
  apply (smt (verit, ccfv_SIG) Store_Rules.st_lv_lc reg_same_st st_opv_daddr_ni st_ots_lm_ni st_ov_ni)
            apply (metis reg_same_st st_opv_daddr_ni)
           apply (metis reg_same_st st_opv_daddr_ni)
          apply (metis fl_opv_daddr_ni reg_same_fl)
         apply (metis fl_opv_daddr_ni reg_same_fl)
        apply (metis reg_same_st st_opv_daddr_ni)
       apply (smt (verit, best) singleton_Un_iff st_opv_daddr_ni st_opv_gen)
      apply (smt (verit, ccfv_threshold) Un_left_absorb insert_is_Un st_opv_daddr_ni st_opv_gen)
     apply (smt (verit, ccfv_threshold) assms(1) fl_opv_daddr_ni fl_opv_ov_rel_s total_wfs_def)
  apply (smt (verit, ccfv_threshold) assms(1) fl_opv_daddr_ni fl_opv_ov_rel_s total_wfs_def)
  using st_opv_daddr_ni apply force
  apply (cases "(pc s) t2"; cases "pc s t1"; simp add: prog_inv_def  prog_def)
  apply (smt (verit) assms(1) cas_nov_daddr_ni cas_opv_ni cas_ots_lm_ni cas_ov_daddr_ni_s cas_possible_lv_lc reg_same_CAS_dt total_wfs_def)
                      apply (metis reg_same_st st_lv_ni st_opv_daddr_ni)
  apply (metis reg_same_st st_lv_ni st_opv_daddr_ni)
                     apply (metis fl_lv_lc fl_opv_daddr_ni reg_same_fl)
                    apply (metis fl_lv_lc fl_opv_daddr_ni reg_same_fl)
  apply (smt (verit, ccfv_threshold) Store_Rules.st_lv_lc reg_same_st st_opv_daddr_ni st_ots_lm_ni st_ov_ni)
  apply (smt (verit, best) Suc_1 cas_fail_lv_ni cas_opv_ni cas_ov_daddr_ni insert_absorb nat.discI reg_same_CAS_dt singleton_insert_inj_eq' subsetD)
  apply (metis reg_same_st st_opv_daddr_ni)
                apply (metis reg_same_st st_opv_daddr_ni)
               apply (metis fl_opv_daddr_ni reg_same_fl)
              apply (metis fl_opv_daddr_ni reg_same_fl)
             apply (metis reg_same_st st_opv_daddr_ni)
  apply (smt (verit, best) assms(1) cas_fail_lv_ni cas_nov_daddr_ni cas_opv_ni cas_ov_daddr_ni_s nat.discI numeral_2_eq_2 total_wfs_def)
  apply (smt (verit, best) assms(1) cas_fail_lv_ni cas_opv_ni cas_ov_daddr_ni_s nat.discI numeral_2_eq_2 reg_same_CAS_dt total_wfs_def)
          apply (metis cas_fail_lv_ni cas_opv_ni zero_neq_numeral)
  apply (smt (verit, ccfv_SIG) assms(1) cas_fail_lv_ni cas_opv_ni cas_ov_daddr_ni_s nat.discI numeral_2_eq_2 total_wfs_def)
  apply (smt (verit, ccfv_SIG) cas_fail_lv_ni cas_opv_ni zero_neq_numeral)
       apply (metis cas_opv_ni)
      apply (smt (verit, best) insert_absorb2 insert_is_Un st_opv_daddr_ni st_opv_gen)
     apply (smt (verit, best) singleton_Un_iff st_opv_daddr_ni st_opv_gen)
  apply (smt (verit, best) OPV_notempty_preserved assms(1) fl_opv_daddr_ni fl_opv_sub subset_singletonD total_wfs_def)
  apply (smt (verit, best) OPV_notempty_preserved assms(1) fl_opv_daddr_ni fl_opv_sub subset_singletonD total_wfs_def)
  using st_opv_daddr_ni by force


(*Initialization*)
theorem init_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "init s"
shows "prog_inv t ((pc s) t) s"
  using assms 
  apply (simp add: init_def start_def prog_inv_def mapval_def OTS_def st_OTSF )   
  using mapval_def st_OPTS st_OPV_z st_lv start_def by force



theorem final_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "t' \<in> {t1, t2}"
  and "prog_inv t ((pc s) t) s"
  and "prog_inv t' ((pc s) t') s" 
  and "prog t' s s'"
shows "prog_inv t ((pc s') t) s'"
  using assms apply (simp del: thrdsvars_def) 
  using t1_global t1_local t2_global t2_local
  by blast


(*Persistence*)
lemma persistence:
  "thrdsvars \<Longrightarrow> prog_inv t2 ((pc s) t2)  ( s) \<Longrightarrow> 
   [z]\<^sub>P (state s) = {1} \<longrightarrow> ([x]\<^sub>P (state s) = {1} \<and>  [y]\<^sub>P (state s) = {1 })"
  apply (simp add: prog_inv_def)
  apply ( cases "pc s t2";   simp add: split if_split )
         apply force
        apply force
       apply force
      apply (metis One_nat_def the_elem_eq zero_neq_one)
     apply metis
    apply metis
   apply (metis One_nat_def the_elem_eq zero_neq_one)
  by (metis One_nat_def the_elem_eq zero_neq_one)
end




