(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

(*The soundness of all the examples is shown with the use of the pierogi proof rules. 
The soundness proof consists of 4 subproofs
 1)Initialisation:    Ensures that the initial assertion of each thread holds at the beginning.
 2)Local Correctness: Establishes annotation validity for each thread.
 3)Stability:         Ensures that each (local)thread annotation is interference-free under the execution of other threads.
 4)Persistence:       Ensures that the crash invariant holds at every program point.*)

(*Epoch Persistency litmus test*)
(*Program outline
  x := 2  |  x := 1 
          |  flush\<^sub>o x
          |  r1 \<leftarrow> x
          |  if (r1 = 2) 
          |     y := 1
          |  sfence
          |  z := 1 
Crash invariant:  (([y]\<^sub>P  = {1} \<and> [z]\<^sub>P  = {1 }) \<longrightarrow>  [x]\<^sub>P  \<subseteq> {1,2}) \<and> [x]\<^sub>P \<noteq> {} *)
 
theory Epoch_Persistency1
imports  "../pierogi/Proof_Rules" 
begin

datatype PC =
   L1
  |L2
  |L3 
  |L4
  |L5
  |L6
  |L7
  |L8



consts t1 ::  TId 
consts t2 ::  TId
consts x :: Loc
consts y :: Loc
consts z :: Loc

consts r1 :: Reg

definition "thrdsvars \<equiv> t1 \<noteq> t2 \<and> 
                         x \<noteq> y  \<and>
                         z \<noteq> y \<and>
                         z \<noteq> x  " 


(*the state of the program*)
record mp_state =
  pc :: "TId \<Rightarrow> PC"
  state :: TState
  aux1 :: nat 
  aux2 :: nat 

(*auxiliary definition for updating the pc*)
definition 
  "update_pc t nv pcf \<equiv> pcf (t := nv)"

lemmas mp_sis [simp] = 
  update_pc_def 
  thrdsvars_def

(*program encoding*)
definition prog :: "TId \<Rightarrow>  mp_state \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog t ts ts' \<equiv> 
  if t = t1
   then
    (case (pc ts) t of 
      L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [x:=2]\<^sub>t (state ts') 
    
   
    | _ \<Rightarrow> ts = ts')
  else 
    if t = t2
    then 
  (case (pc ts) t of 

     L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [x:=1]\<^sub>t (state ts')
    |  L2 \<Rightarrow> pc ts'  = update_pc t  L3 (pc ts) \<and>
      (state ts)  [flush\<^sub>o x ]\<^sub>t (state ts')

    | L3  \<Rightarrow> pc ts' = update_pc t L4 (pc ts) \<and>
            (state ts) [r1   \<leftarrow> x]\<^sub>t (state ts')

   |  L4 \<Rightarrow> if  (regs (state ts) t2  r1  =  2) then
                pc ts' = update_pc t L5 (pc ts) \<and>
                state ts = state ts'
            else pc ts' = update_pc t L6 (pc ts) \<and>
                 state ts = state ts' 
   |  L5 \<Rightarrow> (state ts)   [y:=1]\<^sub>t  (state ts') \<and>
             pc ts' = update_pc t L6 (pc ts)

   |L6 \<Rightarrow>  (state ts)   [sfence]\<^sub>t  (state ts') \<and>
           pc ts' = update_pc t L7 (pc ts)

   |L7 \<Rightarrow>  (state ts)   [z:=1]\<^sub>t  (state ts')\<and>
    pc ts' = update_pc t L8 (pc ts) 
 
  |_ \<Rightarrow> ts = ts'

  )
    else ts = ts'"


(*program invariant*)
definition prog_inv :: "TId \<Rightarrow> PC \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog_inv t p ts \<equiv>
 if t = t1
   then 

   (case p of
          L1 \<Rightarrow>  ( [x]\<^sub>t2  (state ts) = {0}  \<or> ( [x]\<^sub>t2  (state ts) = {1} ))
 
        | L2 \<Rightarrow> True
        | L3 \<Rightarrow> True
        | L4 \<Rightarrow> True
        | L5 \<Rightarrow> True
        | L6 \<Rightarrow> True
        | L7 \<Rightarrow> True
        | L8 \<Rightarrow> True

   
      )
  else
 if t = t2 then
      (case p of
          L1 \<Rightarrow>  [y]\<^sub>P  (state ts) = {0}  \<and>   [z]\<^sub>P (state ts) =  {0} 

        | L2 \<Rightarrow>  ( [x]\<^sub>t (state ts) =  {1} \<or>  [x]\<^sub>t (state ts) =  {1,2}) \<and>    [y]\<^sub>P (state ts) =  {0} \<and>   [z]\<^sub>P (state ts) =  {0} 
     
        | L3 \<Rightarrow>( ( [x]\<^sup>A\<^sub>t (state ts) \<subseteq> {1} \<and>  [x]\<^sub>t (state ts) =  {1}) \<or>
                ( [x]\<^sup>A\<^sub>t (state ts) \<subseteq> {1,2} \<and>  [x]\<^sub>t (state ts) =  {1,2}) )  \<and>    [y]\<^sub>P (state ts) =  {0} \<and>   [z]\<^sub>P (state ts) =  {0}

        |L4 \<Rightarrow> ( (regs (state ts) t2  r1  =  2)   \<longrightarrow>   [x]\<^sup>A\<^sub>t (state ts) \<subseteq> {1,2 } ) \<and>    [y]\<^sub>P (state ts) =  {0}  \<and>   [z]\<^sub>P (state ts) =  {0}  

       | L5 \<Rightarrow>  ( [x]\<^sup>A\<^sub>t (state ts) \<subseteq> {1,2 } )   \<and>    [y]\<^sub>P (state ts) =  {0}  \<and>   [z]\<^sub>P (state ts) =  {0}
   
       | L6 \<Rightarrow> (  [x]\<^sup>A\<^sub>t (state ts) \<subseteq> {1,2 }\<and> [z]\<^sub>P (state ts) =  {0})   \<or> ( [y]\<^sub>P (state ts) =  {0}  \<and>  [z]\<^sub>P (state ts) =  {0})

       | L7 \<Rightarrow>  ( [x]\<^sub>P (state ts) \<subseteq> {1,2 }   \<or>  [y]\<^sub>P (state ts) =  {0} )  \<and>  [z]\<^sub>P (state ts) =  {0}

       | L8 \<Rightarrow>  [x]\<^sub>P (state ts) \<subseteq> {1,2 }   \<or>  [y]\<^sub>P (state ts) =  {0}  \<or>  [z]\<^sub>P (state ts) =  {0}

      )       
  else False " 


(*Initial State*)
definition init  :: "mp_state  \<Rightarrow> bool"
  where
    "init s \<equiv> (pc s) t1 = L1 
      \<and> (pc s) t2 = L1
      \<and> aux1 s  = 0
      \<and> aux2 s = 0
      \<and> start (state s)"


(*Local Correctness of thread 1*)
lemma t1_local:
  assumes "thrdsvars"
      and "total_wfs  (state s)"
      and "prog_inv t1 ((pc s) t1) s" 
      and "prog t1 s s'"
    shows "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  by (cases "pc s t1", simp_all)


(*Local Correctness of thread 2*)
lemma t2_local:
  assumes "thrdsvars"
      and "total_wfs (state s)"
      and "prog_inv t2 ((pc s) t2) s" 
      and "prog t2 s s'"
    shows "prog_inv t2 ((pc s') t2) s'"
 using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t2"; simp add: split if_splits)
       apply (metis st_opv_daddr_ni st_ov_lc)
      apply (smt (verit, best) assms(2) flo_oav_ov_rel flo_opv_ni flo_ov_ni total_wfs_def)
  apply (smt (z3) insert_mono ld_oav_ni ld_opv_ni singleton_insert_inj_eq' subset_singletonD)
    apply (simp split: if_splits)
  apply (metis st_oav_daddr_ni st_opv_daddr_ni)
  apply (smt (verit, best) OPV_notempty_preserved assms(2) le_supI1 opv_oav_sf sf_opv_sub subset_Un_eq subset_singleton_iff total_wfs_def)
   by (metis st_opv_daddr_ni)


(*Stability check for thread 1*)
lemma t1_global :
  assumes  "total_wfs  (state s)"
  and  "prog_inv t1 ((pc s) t1) s"
  and "prog_inv t2 ((pc s) t2) s"
  and "thrdsvars"
  and "prog t2 s s'"
shows  "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_inv_def prog_def total_wfs_def  )
 apply (cases "pc s t1"; cases "pc s t2";  simp add: split if_splits)
  using st_oav_gen st_ov_lc apply force

              apply (smt (z3) fl_oav_sub flo_ov_ni subset_trans)
             apply (smt (verit, ccfv_SIG) assms(1) ld_oav_ni ld_ots_s total_wfs_def)
            apply (simp split:if_splits )
           apply (metis st_ov_ni)
          apply (metis sf_ov_ni)
         apply (metis st_ov_ni)
        apply (simp split:if_splits )
       apply (cases "pc s' t1";simp)
      apply (simp split:if_splits )
      apply (simp split:if_splits )
     apply (simp split:if_splits )
    apply (simp split:if_splits )
   apply (simp split:if_splits )
  by (simp split:if_splits )


(*Stability check for thread 2*)
lemma t2_global :
  assumes  "total_wfs  (state s)"
  and "prog_inv t2 ((pc s) t2) s"
  and "prog_inv t1 ((pc s) t1) s"
  and "thrdsvars"
  and "prog t1 s s'"
shows  "prog_inv t2 ((pc s') t2) s'"
  using assms 
  apply (simp add: prog_inv_def  prog_def total_wfs_def )
  apply (cases "pc s t2"; cases "pc s t1";  simp add: split if_splits)
  apply (metis st_opv_daddr_ni)
        apply (smt (z3) insert_is_Un singleton_insert_inj_eq st_opv_daddr_ni st_ov_dt_lc)
       apply (smt (verit) insert_is_Un set_eq_subset singleton_insert_inj_eq st_oav_gen st_opv_daddr_ni st_ov_dt_lc subset_singletonD)
  apply (smt (z3) Un_subset_iff reg_same_st st_oav_gen st_opv_daddr_ni subset_insertI)
     apply (smt (z3) Un_subset_iff st_oav_gen st_opv_daddr_ni subset_insertI)
  apply (smt (verit, ccfv_SIG) Un_absorb1 Un_insert_right st_oav_gen st_opv_daddr_ni subset_insertI)
   apply (smt (verit, best) Un_subset_iff st_opv_daddr_ni st_opv_gen subset_insertI)
  by (smt (verit, ccfv_SIG) Un_subset_iff st_opv_daddr_ni st_opv_gen subset_insertI)


(*Initialization*)
theorem init_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "init s"
shows "prog_inv t ((pc s) t) s"
  using assms 
  apply (simp add: init_def start_def prog_inv_def mapval_def OTS_def st_OTSF )   
  using st_OPTS start_def by auto


theorem final_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "t' \<in> {t1, t2}"
  and "prog_inv t ((pc s) t) s"
  and "prog_inv t' ((pc s) t') s" 
  and "prog t' s s'"
shows "prog_inv t ((pc s') t) s'"
  using assms apply (simp del: thrdsvars_def) 
  using t1_global t1_local t2_global t2_local
  by blast

(*Persistence*)
lemma persistence:
 "thrdsvars \<Longrightarrow> prog_inv t2 ((pc s) t2)  ( s) \<Longrightarrow>  total_wfs (state s) \<Longrightarrow>
 (( [y]\<^sub>P (state s) = {1 } \<and>  [z]\<^sub>P (state s) = {1 }) \<longrightarrow>  [x]\<^sub>P (state s) \<subseteq> {1,2 }) \<and> [x]\<^sub>P (state s) \<noteq> {}  "
   apply (simp add: prog_inv_def total_wfs_def)
  apply (case_tac "pc s t2"; simp )
   apply (metis numeral_1_eq_Suc_0 the_elem_eq zero_neq_numeral)
  by (metis numeral_1_eq_Suc_0 the_elem_eq zero_neq_numeral)
 
end




