(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

(*The soundness of all the examples is shown with the use of the pierogi proof rules. 
The soundness proof consists of 4 subproofs
 1)Initialisation:    Ensures that the initial assertion of each thread holds at the beginning. 
 2)Local Correctness: Establishes annotation validity for each thread.
 3)Stability:         Ensures that each (local)thread annotation is interference-free under the execution of other threads.
 4)Persistence:       Ensures that the crash invariant holds at every program point.*)

(*Epoch Persistency litmus test*)
(*Program outline
  x := 2  |  x := 1 
          |  r1 \<leftarrow> x 
          |  flush\<^sub>o x
          |  if (r1 = 2) 
          |     y := 1
          |  sfence
          |  z := 1 
Crash invariant:   [y]\<^sub>P  = {1} \<and> [z]\<^sub>P  = {1} \<longrightarrow> [x]\<^sub>P  = {2} *)
 

theory Epoch_Persistency2
imports  "../pierogi/Proof_Rules" 
begin

datatype PC =
   L1
  | L2
  |L3 
  |L4
  |L5
  |L6
  |L7
  |L8



consts t1 ::  TId 
consts t2 ::  TId
consts x :: Loc
consts y :: Loc
consts z :: Loc

consts r1 :: Reg

definition "thrdsvars \<equiv> t1 \<noteq> t2 \<and> 
                         x \<noteq> y  \<and>
                         z \<noteq> y \<and>
                         z \<noteq> x  " 

(*the state of the program*)
record mp_state =
  pc :: "TId \<Rightarrow> PC"
  state :: TState
  aux1 :: nat 
  aux2 :: nat 

(*auxiliary definition for updating the pc*)
definition 
  "update_pc t nv pcf \<equiv> pcf (t := nv)"

lemmas mp_sis [simp] = 
  update_pc_def 
  thrdsvars_def

(*program encoding*)
definition prog :: "TId \<Rightarrow>  mp_state \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog t ts ts' \<equiv> 
  if t = t1
   then
    (case (pc ts) t of 
      L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [x:=2]\<^sub>t (state ts') 
      
    | _ \<Rightarrow> ts = ts')
  else 
    if t = t2
    then 
  (case (pc ts) t of 

     L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [x:=1]\<^sub>t (state ts')
   

    | L2  \<Rightarrow> pc ts' = update_pc t L3 (pc ts) \<and>
            (state ts) [r1   \<leftarrow> x]\<^sub>t (state ts')

   |  L3\<Rightarrow> pc ts'  = update_pc t  L4 (pc ts) \<and>
      (state ts)  [flush\<^sub>o x ]\<^sub>t (state ts')

   |  L4 \<Rightarrow> if  (regs (state ts) t2  r1  =  2) then
                pc ts' = update_pc t L5 (pc ts) \<and>
                state ts = state ts'
            else pc ts' = update_pc t L6 (pc ts) \<and>
                 state ts = state ts' 
   |  L5 \<Rightarrow> (state ts)   [y:=1]\<^sub>t  (state ts') \<and>
             pc ts' = update_pc t L6 (pc ts)

   |L6 \<Rightarrow>  (state ts)   [sfence]\<^sub>t  (state ts') \<and>
           pc ts' = update_pc t L7 (pc ts)

   |L7 \<Rightarrow>  (state ts)   [z:=1]\<^sub>t  (state ts')\<and>
    pc ts' = update_pc t L8 (pc ts) 
 
  |_ \<Rightarrow> ts = ts'

  )
    else ts = ts'"


(*program invariant*)
definition prog_inv :: "TId \<Rightarrow> PC \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog_inv t p ts \<equiv>
 if t = t1
   then 

   (case p of
          L1 \<Rightarrow>  \<lparr>x,2\<rparr>  (state ts) = 0 \<and> (( [x]\<^sub>t2 (state ts) = {0}\<and> [x]\<^sub>t (state ts)  = {0})  \<or> ( [x]\<^sub>t2  (state ts) = {1}\<and> [x]\<^sub>t (state ts)  = {0,1} ))
 
        | L2 \<Rightarrow>  \<lparr>x,2\<rparr>  (state ts) = 1\<and> ( (\<lceil>x,2\<rceil>\<^sub>t  (state ts)\<and> ([x]\<^sub>t2 (state ts) \<subseteq> {1,2} ) ) \<or>  [x]\<^sub>t2 (state ts) \<subseteq> {0,1,2}   )
        | L3 \<Rightarrow> True
        | L4 \<Rightarrow> True
        | L5 \<Rightarrow> True
        | L6 \<Rightarrow> True
        | L7 \<Rightarrow> True
        | L8 \<Rightarrow> True

   
      )
  else
 if t = t2 then
      (case p of
          L1 \<Rightarrow>  [y]\<^sub>P  (state ts) = {0}  \<and>   [z]\<^sub>P (state ts) =  {0} \<and>  \<lparr>x,2\<rparr>  (state ts) \<in> {0,1}

        | L2 \<Rightarrow>  ( [x]\<^sub>t (state ts) =  {1} \<or> ( [x]\<^sub>t (state ts) =  {1,2}\<and>  \<lparr>x,2\<rparr>  (state ts) = 1 \<and> \<lceil>x,2\<rceil>\<^sub>t1  (state ts) ) ) \<and>    [y]\<^sub>P (state ts) =  {0} \<and>   [z]\<^sub>P (state ts) =  {0} 
     
        | L3 \<Rightarrow>( (regs (state ts) t2  r1  =  2)   \<longrightarrow>   [x]\<^sub>t (state ts) = {2} ) \<and>    [y]\<^sub>P (state ts) =  {0}  \<and>   [z]\<^sub>P (state ts) =  {0} 

        |L4 \<Rightarrow> ( (regs (state ts) t2  r1  =  2)   \<longrightarrow>   [x]\<^sup>A\<^sub>t (state ts) = {2} ) \<and>    [y]\<^sub>P (state ts) =  {0}  \<and>   [z]\<^sub>P (state ts) =  {0}  

       | L5 \<Rightarrow>  ( [x]\<^sup>A\<^sub>t (state ts) = {2}  )   \<and>    [y]\<^sub>P (state ts) =  {0}  \<and>   [z]\<^sub>P (state ts) =  {0}
   
       | L6 \<Rightarrow> (  [x]\<^sup>A\<^sub>t (state ts) = {2}\<and> [z]\<^sub>P (state ts) =  {0})   \<or> ( [y]\<^sub>P (state ts) =  {0}  \<and>  [z]\<^sub>P (state ts) =  {0})

       | L7 \<Rightarrow>  ( [x]\<^sub>P (state ts) = { 2 }   \<or>  [y]\<^sub>P (state ts) =  {0} ) 

       | L8 \<Rightarrow>  [x]\<^sub>P (state ts) = { 2 }   \<or>  [y]\<^sub>P (state ts) =  {0} \<or>  [z]\<^sub>P (state ts) =  {0}

      )       
  else False " 


(*Initial State*)
definition init  :: "mp_state  \<Rightarrow> bool"
  where
    "init s \<equiv> (pc s) t1 = L1 
      \<and> (pc s) t2 = L1
      \<and> aux1 s  = 0
      \<and> aux2 s = 0
      \<and> start (state s)"


(*Local Correctness of thread 1*)
lemma t1_local:
  assumes "thrdsvars"
      and "total_wfs  (state s)"
      and "prog_inv t1 ((pc s) t1) s" 
      and "prog t1 s s'"
    shows "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t1", simp_all)
  apply (rule conjI)
   apply (metis Nat.add_0_right add_Suc_right numeral_1_eq_Suc_0 numerals(1) st_oc_lc)
  by (smt (z3) Un_empty_right Un_insert_right insert_commute st_ov_dt_lc subset_insertI)

  
(*Local Correctness of thread 2*)
lemma t2_local:
  assumes "thrdsvars"
      and "total_wfs (state s)"
      and "prog_inv t2 ((pc s) t2) s" 
      and "prog t2 s s'"
    shows "prog_inv t2 ((pc s') t2) s'"
 using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t2"; simp add: split if_splits)
       apply (metis st_opv_daddr_ni st_ov_lc)
      apply (intro conjI)
  apply (smt (verit, ccfv_SIG) assms(2) insert_absorb insert_iff insert_not_empty ld_ov_lc ld_ov_oc_lm_lc numeral_1_eq_Suc_0 numeral_eq_iff numerals(1) semiring_norm(83) total_wfs_def)
  apply (metis ld_opv_ni)
      apply (metis ld_opv_ni)
  apply (metis assms(2) flo_oav_ov_rel_s flo_opv_ni reg_same_flo total_wfs_def)
    apply (simp split:if_splits )
   apply (metis st_oav_daddr_ni st_opv_daddr_ni)
 apply (metis OPV_notempty_preserved assms(2) opv_oav_sf sf_opv_sub subset_singleton_iff total_wfs_def)
  using st_opv_daddr_ni by force


 (*Stability check for thread 1*)
lemma t1_global :
  assumes  "total_wfs  (state s)"
  and  "prog_inv t1 ((pc s) t1) s"
  and "prog_inv t2 ((pc s) t2) s"
  and "thrdsvars"
  and "prog t2 s s'"
shows  "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_inv_def prog_def total_wfs_def  )
 apply (cases "pc s t1"; cases "pc s t2";  simp add: split if_splits)
                     apply (intro conjI)
  apply (metis numeral_1_eq_Suc_0 numeral_eq_iff semiring_norm(83) st_noc_lc)
                     apply (smt (z3) insert_is_Un st_ov_dt_lc st_ov_lc sup.right_idem)
                    apply (metis (no_types, lifting) OV_notempty_preserved assms(1) ld_oc_ni ld_ov_ni ld_ov_sub subset_singletonD total_wfs_def)
  using flo_oc_ni flo_ov_ni apply force
                  apply (simp split:if_splits )
                 apply (elim conjE disjE)
  apply (metis n_not_Suc_n numeral_2_eq_2 st_noc_lc st_ov_ni)
                 apply (simp add: st_noc_lc st_ov_ni)
  using sf_oc_ni sf_ov_ni apply force
  using st_noc_lc st_ov_ni apply force
  apply (metis insert_subset n_not_Suc_n numeral_2_eq_2 st_noc_lc st_ov_lc subset_insertI)
             apply (smt (verit, ccfv_SIG) insert_commute ld_oc_ni ld_ov_sub subset_insertI2)
  apply (metis (no_types, lifting) flo_oc_ni flo_ov_ni subset_insertI2)
           apply (simp split:if_splits )
          apply (metis card_1_singleton_iff card_2_iff' singleton_iff st_noc_lc st_ov_ni subset_insertI2)
         apply (metis (no_types, lifting) sf_oc_ni sf_ov_ni subset_insertI2)
  apply (metis card_1_singleton_iff card_2_iff' singleton_iff st_noc_lc st_ov_ni subset_insertI2)
       apply (simp split:if_splits )
      apply (simp split:if_splits )
     apply (simp split:if_splits )
    apply (simp split:if_splits )
   apply (simp split:if_splits )
  by (simp split:if_splits )


(*Stability check for thread 2*)
lemma t2_global :
  assumes  "total_wfs  (state s)"
  and "prog_inv t2 ((pc s) t2) s"
  and "prog_inv t1 ((pc s) t1) s"
  and "thrdsvars"
  and "prog t1 s s'"
shows  "prog_inv t2 ((pc s') t2) s'"
  using assms 
  apply (simp add: prog_inv_def  prog_def total_wfs_def )
  apply (cases "pc s t2"; cases "pc s t1";  simp add: split if_splits)
  apply (smt (z3) One_nat_def PC.simps(58) assms(1) assms(3) assms(5) fun_upd_same prog_inv_def st_opv_daddr_ni t1_local thrdsvars_def)
        apply (smt (z3) One_nat_def PC.simps(58) assms(1) assms(3) assms(5) fun_upd_same insert_is_Un prog_inv_def st_opv_daddr_ni st_ots_lm_lc st_ov_dt_lc st_ov_lc t1_local thrdsvars_def vbounded_preserved)
  apply (metis Zero_not_Suc ex_in_conv insert_iff n_not_Suc_n numeral_2_eq_2 reg_same_st st_opv_daddr_ni)
  apply (metis (no_types, lifting) reg_same_st st_oav_gen st_opv_daddr_ni sup.idem)  
     apply (metis Un_left_absorb insert_is_Un st_oav_gen st_opv_daddr_ni)
  apply (metis insert_absorb2 insert_is_Un st_oav_gen st_opv_daddr_ni)
  using st_opv_daddr_ni st_opv_gen apply force
  by (metis (mono_tags) Un_empty_right Un_insert_right insert_absorb2 st_opv_daddr_ni st_opv_gen)

(*Initialization*)
theorem init_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "init s"
shows "prog_inv t ((pc s) t) s"
  using assms 
  apply (simp add: init_def start_def prog_inv_def mapval_def OTS_def st_OTSF ) 
  using assms(4) init_def st_OPTS st_oc by force
  

theorem final_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "t' \<in> {t1, t2}"
  and "prog_inv t ((pc s) t) s"
  and "prog_inv t' ((pc s) t') s" 
  and "prog t' s s'"
shows "prog_inv t ((pc s') t) s'"
  using assms apply (simp del: thrdsvars_def) 
  using t1_global t1_local t2_global t2_local
  by blast

(*Persistence*)
lemma persistence:
 "thrdsvars \<Longrightarrow> prog_inv t2 ((pc s) t2) s \<Longrightarrow> 
  [y]\<^sub>P (state s) = {1} \<and> [z]\<^sub>P (state s) = {1} \<longrightarrow> [x]\<^sub>P (state s) = {2}"
  apply (simp add: prog_inv_def split:if_splits)
  by (case_tac "pc s t2", auto )





end


