(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

(*The soundness of all the examples is shown with the use of the pierogi proof rules. 
The soundness proof consists of 4 subproofs
 1)Initialisation:    Ensures that the initial assertion of each thread holds at the beginning. 
 2)Local Correctness: Establishes annotation validity for each thread.
 3)Stability:         Ensures that each (local)thread annotation is interference-free under the execution of other threads.
 4)Persistence:       Ensures that the crash invariant holds at every program point.*)

(*Buffering with flushopts litmus test*)
(*Program outline
  x := 1       |   y := 1 
  flush y      |   flush x
  w := 1       |   z := 1 
Crash invariant:    [z]\<^sub>P = {1} \<and>  [w]\<^sub>P  = {1} \<longrightarrow> ([x]\<^sub>P = {1} \<or>  [y]\<^sub>P = {1}) *)
 


theory Flush_Buffering
imports  "../pierogi/Proof_Rules" 
begin

datatype PC =
  L1
| L2
| L3
| L4

consts t1 ::  TId 
consts t2 ::  TId
consts x :: Loc
consts y :: Loc
consts w :: Loc
consts z :: Loc

definition "thrdsvars \<equiv> t1 \<noteq> t2 \<and> 
                         x \<noteq> y  \<and>
                         w \<noteq> y \<and>
                         w \<noteq> x  \<and>
                         z \<noteq> x \<and>
                         z \<noteq> y \<and>
                         z \<noteq> w  " 


(*the state of the program*)
record mp_state =
  pc :: "TId \<Rightarrow> PC"
  state :: TState
  aux1 :: nat 
  aux2 :: nat 

(*auxiliary definition for updating the pc*)
definition 
  "update_pc t nv pcf \<equiv> pcf (t := nv)"

lemmas mp_sis [simp] = 
  update_pc_def 
  thrdsvars_def

(*program encoding*)
definition prog :: "TId \<Rightarrow>  mp_state \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog t ts ts' \<equiv> 
  if t = t1
   then
    (case (pc ts) t of 
      L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [x:=1]\<^sub>t (state ts') \<and>
      (aux1 ts') = ( aux2 ts)  + 1  \<and>
      ( aux2 ts')  =  ( aux2 ts) 
    | L2 \<Rightarrow> pc ts' = update_pc t L3 (pc ts) \<and>
     (state ts)  [flush y]\<^sub>t  (state ts') \<and>
     aux1 ts = aux1 ts' \<and>
     aux2 ts = aux2  ts'
   | L3 \<Rightarrow>  pc ts' = update_pc t L4 (pc ts) \<and>
       (state ts) [w:=1]\<^sub>t (state ts') \<and>
       aux1 ts = aux1 ts' \<and>
       aux2 ts = aux2  ts'
    | _ \<Rightarrow> ts = ts')
  else 
    if t = t2
    then 
  (case (pc ts) t of 

     L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [y:=1]\<^sub>t (state ts') \<and>
      (aux2 ts') = ( aux1 ts)  + 1  \<and>
      ( aux1 ts')  =  ( aux1 ts) 
 
     | L2 \<Rightarrow>  pc ts' = update_pc t L3 (pc ts) \<and>
               (state ts)  [flush x]\<^sub>t  (state ts') \<and>
                aux1 ts = aux1 ts' \<and>
                 aux2 ts = aux2  ts' 
     | L3 \<Rightarrow> pc ts' = update_pc t L4 (pc ts) \<and>
       (state ts) [z:=1]\<^sub>t (state ts') \<and>
       aux1 ts = aux1 ts' \<and>
       aux2 ts = aux2  ts'
    |_ \<Rightarrow> ts = ts'

  )
    else ts = ts'"


(*program invariant*)
definition prog_inv :: "TId \<Rightarrow> PC \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog_inv t p ts \<equiv>
 if t = t1
   then 

   (case p of
          L1 \<Rightarrow>( aux1 ts = 0 \<and> aux2 ts = 0 \<and> [z]\<^sub>P (state ts) =  {0}) 
              \<or> ( aux1 ts = 0 \<and> aux2 ts = 1  \<and> \<lceil>y,1\<rceil>\<^sub>t2 (state ts)  \<and> [w]\<^sub>P (state ts) =  {0})
 
        | L2 \<Rightarrow> 
                ( aux1 ts = 2 \<and> aux2 ts = 1 \<and> \<lceil>y,1\<rceil>\<^sub>t2 (state ts) \<and> [w]\<^sub>P (state ts) =  {0}  \<and> \<lceil>FLUSH y\<rceil>\<^sub>t1  (state ts) ) \<or>
                 ( aux1 ts = 1 \<and> (aux2 ts = 0 \<or> aux2 ts = 2) \<and> ( [x]\<^sub>P (state ts) =  {1}  \<or> [z]\<^sub>P (state ts) =  {0}) )

        | L3 \<Rightarrow>  
                ( aux1 ts = 2 \<and> aux2 ts = 1\<and> [y]\<^sub>P (state ts) =  {1} ) \<or>
                 ( aux1 ts = 1 \<and> (aux2 ts = 0 \<or> aux2 ts = 2) \<and>( [x]\<^sub>P (state ts) =  {1}  \<or> [z]\<^sub>P (state ts) =  {0}) )

        | L4 \<Rightarrow>  
                ( aux1 ts = 2 \<and> aux2 ts = 1\<and> [y]\<^sub>P (state ts) =  {1}  ) \<or>
                 ( aux1 ts = 1 \<and>(aux2 ts = 0 \<or> aux2 ts = 2) \<and>( [x]\<^sub>P (state ts) =  {1}  \<or> [z]\<^sub>P (state ts) =  {0}) )
      )
  else
 if t = t2 then
      (case p of
          L1 \<Rightarrow>( aux2 ts = 0 \<and> aux1 ts = 0 \<and> [w]\<^sub>P (state ts) =  {0} ) 
              \<or> ( aux2 ts = 0 \<and> aux1 ts = 1 \<and> \<lceil>x,1\<rceil>\<^sub>t1 (state ts) \<and> [z]\<^sub>P (state ts) =  {0})

        | L2 \<Rightarrow>   
                ( aux2 ts = 2 \<and> aux1 ts = 1 \<and> \<lceil>x,1\<rceil>\<^sub>t1 (state ts) \<and> [z]\<^sub>P (state ts) =  {0}\<and>  \<lceil>FLUSH x\<rceil>\<^sub>t2  (state ts) ) \<or>
                 ( aux2 ts = 1 \<and> (aux1 ts = 2 \<or> aux1 ts = 0) \<and> ( [y]\<^sub>P (state ts) =  {1}  \<or> [w]\<^sub>P (state ts) =  {0}) )

        | L3 \<Rightarrow>  
                ( aux2 ts = 2 \<and> aux1 ts = 1\<and> [x]\<^sub>P (state ts) =  {1} ) \<or>
                 ( aux2 ts = 1 \<and> (aux1 ts = 2 \<or> aux1 ts = 0) \<and> ( [y]\<^sub>P (state ts) =  {1}  \<or> [w]\<^sub>P (state ts) =  {0}) )

        | L4 \<Rightarrow>   
                ( aux2 ts = 2 \<and> aux1 ts = 1\<and> [x]\<^sub>P (state ts) =  {1} ) \<or>
                 ( aux2 ts = 1 \<and> (aux1 ts = 2 \<or> aux1 ts = 0) \<and> ( [y]\<^sub>P (state ts) =  {1}  \<or> [w]\<^sub>P (state ts) =  {0}) )
      )       
  else False " 





(*Initial State*)
definition init  :: "mp_state  \<Rightarrow> bool"
  where
    "init s \<equiv> (pc s) t1 = L1 
      \<and> (pc s) t2 = L1
      \<and> aux1 s  = 0
      \<and> aux2 s = 0
      \<and> start (state s)"


(*Local Correctness of thread 1*)
lemma t1_local:
  assumes "thrdsvars"
      and "total_wfs  (state s)"
      and "prog_inv t1 ((pc s) t1) s" 
      and "prog t1 s s'"
    shows "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t1", simp_all)
    apply (smt (z3) st_opv_daddr_ni st_ord_lc st_ots_lm_ni st_ov_ni) 
 apply (elim  disjE)  
  apply (metis assms(2) fl_ov_opv_ord total_wfs_def)
    apply simp
   apply (smt (verit, best) assms(2) fl_opv_daddr_ni fl_ov_opv_ord total_wfs_def)
  by (metis (no_types, lifting) st_opv_daddr_ni)
   
(*Local Correctness of thread 2*)
lemma t2_local:
  assumes "thrdsvars"
      and "total_wfs (state s)"
      and "prog_inv t2 ((pc s) t2) s" 
      and "prog t2 s s'"
    shows "prog_inv t2 ((pc s') t2) s'"
 using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t2", simp_all)
   apply (elim  disjE)
     apply simp 
     apply (metis st_opv_daddr_ni)
  apply (metis st_opv_daddr_ni st_ord_lc st_ots_lm_ni st_ov_ni)
  apply (elim  disjE)
  apply (metis assms(2) fl_ov_opv_ord total_wfs_def)
  apply (metis fl_opv_daddr_ni)
  using st_opv_daddr_ni by force


(*Stability check for thread 1*)
lemma t1_global :
  assumes  "total_wfs  (state s)"
  and  "prog_inv t1 ((pc s) t1) s"
  and "prog_inv t2 ((pc s) t2) s"
  and "thrdsvars"
  and "prog t2 s s'"
shows  "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_inv_def prog_def total_wfs_def  )
  apply (simp split:if_splits PC.split)
  apply (cases "(pc s) t1", simp_all)
     apply (cases "(pc s) t2", simp_all)
  apply (metis nat.distinct(1) st_opv_daddr_ni st_ots_lm_lc st_ov_lc vbounded_preserved)
  apply (smt (z3) fl_opv_daddr_ni fl_ots_lm_ni fl_ov_ni)
  apply (metis numeral_1_eq_Suc_0 st_opv_daddr_ni st_ots_lm_ni st_ov_ni zero_neq_numeral)
    apply (cases "(pc s) t2", simp_all)
      apply (metis n_not_Suc_n st_opv_daddr_ni)
  apply (smt (z3) fl_opv_daddr_ni fl_ord_ni fl_ots_lm_ni fl_ov_ni n_not_Suc_n numeral_2_eq_2)
  apply (smt (z3) One_nat_def Suc_1 nat.inject st_opv_daddr_ni st_ord_ni st_ots_lm_ni st_ov_ni zero_neq_numeral)
   apply (cases "(pc s) t2", simp_all)
     apply (metis st_opv_daddr_ni zero_neq_numeral)
    apply (metis fl_opv_daddr_ni n_not_Suc_n numeral_2_eq_2)
  apply (metis One_nat_def Suc_1 n_not_Suc_n st_opv_daddr_ni)
  apply (cases "(pc s) t2", simp_all)
    apply (metis st_opv_daddr_ni zero_neq_numeral)
   apply (metis fl_opv_daddr_ni n_not_Suc_n numeral_2_eq_2)
  by (metis One_nat_def Suc_1 n_not_Suc_n st_opv_daddr_ni)


(*Stability check for thread 2*)
lemma t2_global :
  assumes  "total_wfs  (state s)"
  and "prog_inv t2 ((pc s) t2) s"
  and "prog_inv t1 ((pc s) t1) s"
  and "thrdsvars"
  and "prog t1 s s'"
shows  "prog_inv t2 ((pc s') t2) s'"
  using assms 
  apply (simp add: prog_inv_def  prog_def total_wfs_def )
  apply (simp split:if_splits PC.split)
  apply (cases "(pc s) t2", simp_all)
     apply (cases "(pc s) t1", simp_all)
       apply (metis n_not_Suc_n st_opv_daddr_ni st_ots_lm_lc st_ov_lc vbounded_preserved)
      apply (smt (z3) fl_opv_daddr_ni fl_ots_lm_ni fl_ov_ni)
  apply (metis numeral_1_eq_Suc_0 numerals(1) st_opv_daddr_ni st_ots_lm_ni st_ov_ni zero_neq_one)
    apply (cases "(pc s) t1", simp_all)
  apply (metis One_nat_def le_numeral_extra(4) not_one_le_zero st_opv_daddr_ni)
  apply (smt (z3) fl_opv_daddr_ni fl_ord_ni fl_ots_lm_ni fl_ov_ni n_not_Suc_n numeral_2_eq_2)
  apply (metis (no_types, lifting) One_nat_def Suc_1 n_not_Suc_n st_opv_daddr_ni st_ord_ni st_ots_lm_ni st_ov_ni)
   apply (cases "(pc s) t1", simp_all)
  apply (metis Zero_not_Suc st_opv_daddr_ni)
    apply (metis fl_opv_daddr_ni n_not_Suc_n numeral_2_eq_2)
   apply (metis n_not_Suc_n numeral_2_eq_2 st_opv_daddr_ni)
  apply (cases "(pc s) t1", simp_all)
  apply (metis Zero_not_Suc st_opv_daddr_ni)
   apply (metis fl_opv_daddr_ni n_not_Suc_n numeral_2_eq_2)
  by (metis numeral_1_eq_Suc_0 numeral_eq_one_iff semiring_norm(83) st_opv_daddr_ni zero_neq_numeral)

(*Initialization*)
theorem init_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "init s"
shows "prog_inv t ((pc s) t) s"
  using assms 
  apply (simp add: init_def start_def prog_inv_def mapval_def OTS_def st_OTSF )   
  using st_OPTS start_def by auto


theorem final_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "t' \<in> {t1, t2}"
  and "prog_inv t ((pc s) t) s"
  and "prog_inv t' ((pc s) t') s" 
  and "prog t' s s'"
shows "prog_inv t ((pc s') t) s'"
  using assms apply (simp del: thrdsvars_def) 
  using t1_global t1_local t2_global t2_local
  by blast

(*Persistence*)
lemma persistence:
 "thrdsvars \<Longrightarrow> prog_inv t2 ((pc s) t2)  ( s) \<Longrightarrow> 
  [z]\<^sub>P (state s) = {1 } \<and>  [w]\<^sub>P (state s) = {1 } \<longrightarrow> ([x]\<^sub>P (state s) = {1 } \<or>  [y]\<^sub>P (state s) = {1 })"
   apply (simp add: prog_inv_def)
  by (case_tac "pc s t2", auto )

end




