(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

(*The soundness of all the examples is shown with the use of the pierogi proof rules. 
This soundness proof consists of 3 subproofs
 1)Initialisation:    Ensures that the initial assertion of each thread holds at the beginning.
 2)Local Correctness: Establishes annotation validity for each thread.
 3)Stability:         Ensures that each (local)thread annotation is interference-free under the execution of other threads.*)

(*Message Passing litmus test*)
(*Program outline 
 data := 42  | r1   \<leftarrow> flag 
  flag := 7  | if (r1 = 7)
             |    r2   \<leftarrow> data 
*)


theory Message_Passing
imports   "../pierogi/Proof_Rules"
begin
                          
datatype PC = L1 | L2 | L3


consts t1 ::  TId 
consts t2 ::  TId
consts data :: Loc
consts flag :: Loc
consts r1 :: Reg
consts r2 :: Reg 

definition "thrdsvars \<equiv> t1 \<noteq> t2 \<and> 
                        data \<noteq> flag"

(*the state of the program*)
record mp_state =
  pc :: "TId \<Rightarrow> PC"
  state :: TState

(*auxiliary definition for updating the pc*)
definition 
  "update_pc t nv pcf \<equiv> pcf (t := nv)"


(*program encoding*)
definition prog :: "TId \<Rightarrow>  mp_state \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog t ts ts' \<equiv> 
  if t = t1
   then
    (case (pc ts) t of 
      L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
       (state ts) [data:=42]\<^sub>t (state ts') 
    | L2 \<Rightarrow> pc ts' = update_pc t L3 (pc ts) \<and>
              (state ts)  [flag:=7]\<^sub>t  (state ts')
    | _ \<Rightarrow> ts = ts')

    else 
   if t = t2
    then (case (pc ts) t of 
       L1 \<Rightarrow> pc ts' = update_pc t L2 (pc ts) \<and>
            (state ts) [r1   \<leftarrow> flag ]\<^sub>t (state ts') 
    |  L2 \<Rightarrow> if  (regs (state ts) t2  r1  =  7) then
                pc ts' = update_pc t L3 (pc ts) \<and>
                (state ts)  [r2  \<leftarrow> data ]\<^sub>t   (state ts')            
            else pc ts' = update_pc t L3 (pc ts) \<and>
                 state ts = state ts'
    |  _ \<Rightarrow> ts = ts')
    else ts = ts'"

(*program invariant*)
definition prog_inv :: "TId \<Rightarrow> PC \<Rightarrow> TState \<Rightarrow> bool " where
"prog_inv t p ts \<equiv>  
 if t = t1
   then 
   (case p of
          L1 \<Rightarrow> 7 \<notin> [flag]\<^sub>t2 ts \<and> regs ts t2  r1  = 0 
        | L2 \<Rightarrow> [data]\<^sub>t  ts = { 42 } \<and>  7 \<notin> [flag]\<^sub>t2  ts
        | L3 \<Rightarrow> True
      )
  else
 if t = t2 then
      (case p of 
          L1 \<Rightarrow> (7 \<in> [flag]\<^sub>t2 ts \<longrightarrow> \<langle>flag  7\<rangle>[data]\<^sub>t2 ts = {42}) \<and> [flag]\<^sub>t2 ts \<subseteq> { 0,7}
        | L2 \<Rightarrow>  ( (regs ts t2  r1  = 7) \<longrightarrow> [data]\<^sub>t2 ts = {42} \<and>  regs ts t2  r1  \<in> { 0,7} )
        | L3 \<Rightarrow> (regs ts t2  r1  = 7) \<longrightarrow> regs ts t2  r2  = 42 
      )
  else False "


lemmas mp_sis [simp] = 
  update_pc_def 
  thrdsvars_def

(*Initial State*)
definition init  :: "mp_state  \<Rightarrow> bool"
  where
    "init s \<equiv> (pc s) t1 = L1 
      \<and> (pc s) t2 = L1
    \<and> start (state s)"


(*Initialization*)
theorem init_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "init s"
shows "prog_inv t ((pc s) t) (state s)"
  using assms 
  apply (simp add: init_def start_def prog_inv_def )
  by (metis bot.extremum emptyE insert_iff insert_subset st_OV_z start_def zero_neq_numeral)


(*Local Correctness of thread 1*)
lemma t1_local:
  assumes "thrdsvars"
      and "total_wfs (state s)"
        and "prog_inv t1 ((pc s) t1) (state s)" 
      and "prog t1  s  s'"
        shows "prog_inv t1 ((pc s') t1) (state s')"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t1", simp_all)
  using st_ov_lc st_ov_ni by blast

 
(*Local Correctness of thread 2*)
lemma t2_local:
  assumes "thrdsvars"
      and "total_wfs (state s)"
      and "prog_inv t2 ((pc s) t2) (state s)" 
      and "prog t2 s s'"
    shows "prog_inv t2 ((pc s') t2) (state s')"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
apply (cases "pc s t2", simp_all split: if_splits)
  using assms(2) ld_cobv_lc_s total_wfs_def 
   apply (metis ld_ov_lc) 
  by (metis ld_ov_lc singletonD)

  
(*Stability check for thread 1*)
lemma t1_global :
  assumes  "total_wfs  (state s)"
  and  "prog_inv t1 ((pc s) t1)  (state s)"
  and "prog_inv t2 ((pc s) t2)  (state s)"
  and "thrdsvars"
  and "prog t2 s s'"
shows  "prog_inv t1 ((pc s') t1)  (state s')"
  using assms 
  apply (simp add: prog_inv_def prog_def total_wfs_def )
  apply (simp split: if_splits)
  apply (cases "pc s t2"; cases "pc s t1";   simp add: split if_split )
    apply (metis ld_nov_ni ld_ov_ni)
   apply (metis ld_nov_ni ld_ov_ni)
  apply (cases "pc s t1"; cases "pc s t2";   simp add: split if_split )
  apply (smt (z3) in_mono insert_commute ld_nov_ni ld_ov_lc singleton_iff subset_insert)
  by (metis ld_nov_ni ld_ov_ni)

(*Stability check for thread 2*)
lemma t2_global :
  assumes  "total_wfs  (state s)"
  and "prog_inv t2 ((pc s) t2)  (state s)"
  and "prog_inv t1 ((pc s) t1)  (state s)"
  and "thrdsvars"
  and "prog t1 s s'"
shows  "prog_inv t2 ((pc s') t2)  (state s')"
  using assms 
  apply (cases "pc s t1"; cases "pc s t2";  simp add: prog_inv_def  prog_def )
  using total_wfs_def  st_ov_ni apply blast
  using total_wfs_def  apply (metis reg_same_st zero_neq_numeral)
     apply (metis reg_same_st zero_neq_numeral)
  using total_wfs_def   
  apply (smt (verit, ccfv_SIG) Un_least ld_cobv_lc_intro_s st_ov_dt_lc subset_insertI total_OTSF_def)
    using total_wfs_def 
     apply (metis reg_same_st st_ov_ni)
  by (metis reg_same_st)


theorem final_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "t' \<in> {t1, t2}"
  and "prog_inv t ((pc s) t) (state s)"
  and "prog_inv t' ((pc s) t') (state s)" 
  and "prog t' s s'"
shows "prog_inv t ((pc s') t) (state s')"
  using assms apply (simp del: thrdsvars_def) 
  using t1_global t1_local t2_global t2_local by blast



end 

