(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

(*The soundness of all the examples is shown with the use of the pierogi proof rules. 
The soundness proof consists of 4 subproofs
 1)Initialisation:    Ensures that the initial assertion of each thread holds at the beginning.
 2)Local Correctness: Establishes annotation validity for each thread.
 3)Stability:         Ensures that each (local)thread annotation is interference-free under the execution of other threads.
 4)Persistence:       Ensures that the crash invariant holds at every program point.*)

(*Message Passing with flushes litmus test*)
(*Program outline 
  data := 4  | r1   \<leftarrow> flag 
  flush data | if (r1 = 7)
  flag := 7  |    commit := 1
Crash invariant:  [commit]\<^sub>P = {1}  \<longrightarrow> [data]\<^sub>P = {42}
*)

theory Persistent_Message_Passing2
imports   "../pierogi/Proof_Rules"
begin

datatype PC =
  L1
| L2
| L3
| L4

consts t1 ::  TId 
consts t2 ::  TId
consts data :: Loc
consts flag :: Loc
consts commit :: Loc
consts r1 :: Reg
consts r2 :: Reg 



definition "thrdsvars \<equiv> t1 \<noteq> t2 \<and> 
                        data \<noteq> flag  \<and>
                        commit \<noteq> flag \<and>
                        commit \<noteq> data" 

(*the state of the program*)
record mp_state =
  pc :: "TId \<Rightarrow> PC"
  state :: TState


(*auxiliary definition for updating the pc*)
definition 
  "update_pc t nv pcf \<equiv> pcf (t := nv)"

lemmas mp_sis [simp] = 
  update_pc_def 
  thrdsvars_def

(*program encoding*)
definition prog :: "TId \<Rightarrow>  mp_state \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog t ts ts' \<equiv> 
  if t = t1
   then
    (case (pc ts) t of 
      L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
       (state ts) [data:=42]\<^sub>t (state ts') 
    | L2 \<Rightarrow> pc ts' = update_pc t L3 (pc ts) \<and>
           (state ts)   [flush data ]\<^sub>t  (state ts')
   | L3 \<Rightarrow> pc ts' = update_pc t L4 (pc ts) \<and>
             (state ts)  [flag := 7]\<^sub>t  (state ts')
    | _ \<Rightarrow> ts = ts')

    else 
    if t = t2
    then (case (pc ts) t of 
       L1 \<Rightarrow> pc ts' = update_pc t L2 (pc ts) \<and>
            (state ts) [r1   \<leftarrow> flag ]\<^sub>t (state ts') 
    |  L2 \<Rightarrow> if  (regs (state ts) t2  r1  =  7) then
                pc ts' = update_pc t L3 (pc ts) \<and>
                state ts = state ts'
            else pc ts' = update_pc t L4 (pc ts) \<and>
                 state ts = state ts' 
    |  L3 \<Rightarrow> (state ts)   [commit:=1]\<^sub>t  (state ts') \<and>
             pc ts' = update_pc t L4 (pc ts)
    | _ \<Rightarrow> ts = ts')
    else ts = ts'"


(*program invariant*)
definition prog_inv :: "TId \<Rightarrow> PC \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog_inv t p ts \<equiv> 
 if t = t1
   then 

   (case p of
          L1 \<Rightarrow>  [commit]\<^sub>P (state ts) = {0} \<and> [flag]\<^sub>t2 (state ts) = {0} \<and> (regs (state ts) t2  r1  =  0)
        | L2 \<Rightarrow>  [commit]\<^sub>P (state ts) = {0} \<and> [flag]\<^sub>t2 (state ts) = {0} \<and> (regs (state ts) t2  r1  =  0) \<and> [data]\<^sub>t1 (state ts) = {42}
        | L3 \<Rightarrow>  [data]\<^sub>P (state ts) = {42} 
        | L4 \<Rightarrow>  [data]\<^sub>P (state ts) = {42}
      )
  else
    if t = t2 then 
     (case p of
          L1 \<Rightarrow>  True
        | L2 \<Rightarrow>  True
        | L3 \<Rightarrow>  (regs (state ts) t2  r1  =  7)
        | L4 \<Rightarrow>  True)
  else False "

(*Initial State*)
definition init  :: "mp_state  \<Rightarrow> bool"
  where
    "init s \<equiv> (pc s) t1 = L1 
      \<and> (pc s) t2 = L1
    \<and> start (state s)"


(*Local Correctness of thread 1*)
lemma t1_local:
  assumes "thrdsvars"
      and "total_wfs  (state s)"
      and "prog_inv t1 ((pc s) t1) s" 
      and "prog t1 s s'"    
    shows "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t1", simp_all)
    apply (meson st_ov_lc)
  apply (metis st_opv_daddr_ni)
     apply (metis st_ov_ni)
  apply (metis reg_same_st)
   apply (meson assms(2) fl_opv_ov_rel_s total_wfs_def)
  by (metis st_opv_daddr_ni)

(*Local Correctness of thread 2*)
lemma t2_local:
  assumes "thrdsvars"
      and "total_wfs (state s)"
      and "prog_inv t2 ((pc s) t2) s" 
      and "prog t2 s s'"
    shows "prog_inv t2 ((pc s') t2) s'"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def )
  apply (cases "pc s t2", simp_all) 
  by (simp split: if_splits)



(*Stability check for thread 1*)
lemma t1_global :
  assumes  "total_wfs  (state s)"
  and  "prog_inv t1 ((pc s) t1) s"
  and "prog_inv t2 ((pc s) t2) s"
  and "thrdsvars"
  and "prog t2 s s'"
 
shows  "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_inv_def prog_def total_wfs_def )
  apply (cases "pc s t1"; cases "pc s t2"; simp)
         apply (simp split: if_splits)
  apply (metis OV_notempty_preserved assms(1) ld_opv_ni ld_ov_lc ld_ov_sub singletonD subset_singletonD total_wfs_def)
  apply (metis OV_notempty_preserved assms(1) empty_iff insert_iff ld_opv_ni ld_ov_lc ld_ov_ni ld_ov_sub subset_singletonD total_wfs_def)
  apply (metis OV_notempty_preserved assms(1) empty_iff insert_iff ld_opv_ni ld_ov_lc ld_ov_sub subset_singletonD total_wfs_def)

  apply (simp split: if_splits)
  using st_opv_daddr_ni apply blast
  apply (metis ld_opv_ni)
  apply (smt (z3) PC.simps(16) fun_upd_apply mp_sis(1))
  by (metis st_opv_daddr_ni)


 
(*Stability check for thread 2*)
lemma t2_global :
  assumes  "total_wfs  (state s)"
  and "prog_inv t2 ((pc s) t2) s"
  and "prog_inv t1 ((pc s) t1) s"
  and "thrdsvars"
  and "prog t1 s s'"
  
shows  "prog_inv t2 ((pc s') t2) s'"
  using assms 
   apply (simp add: prog_inv_def  prog_def total_wfs_def  )
  apply (cases "(pc s) t2"; cases "(pc s) t1"; simp add: split if_splits)
  by (metis reg_same_st)


(*Initialization*)
theorem init_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "init s"
 
shows "prog_inv t ((pc s) t) s"
  using assms 
  apply (simp add: init_def start_def prog_inv_def )
  using imageE image_insert image_is_empty insert_commute mapval_def nth_Cons_0 st_OPTS st_OTS start_def subset_insertI zero_neq_numeral by auto



theorem final_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "t' \<in> {t1, t2}"
  and "prog_inv t ((pc s) t) s"
  and "prog_inv t' ((pc s) t') s" 
  and "prog t' s s'"
shows "prog_inv t ((pc s') t) s'"
  using assms apply (simp del: thrdsvars_def) 
  using t1_global t1_local t2_global t2_local by blast

(*Persistence*)
lemma persistence:
 "thrdsvars \<Longrightarrow> prog_inv t1 ((pc s) t1)  ( s) \<Longrightarrow> 
  [commit]\<^sub>P (state s) = {1 } \<longrightarrow> ([data]\<^sub>P (state s) = {42 } )"
    apply (simp add: prog_inv_def split:if_splits)
  by (case_tac "pc s t1";simp )
 
  


end




