(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

(*The soundness of all the examples is shown with the use of the pierogi proof rules. 
This soundness proof consists of 3 subproofs
 1)Initialisation:    Ensures that the initial assertion of each thread holds at the beginning.
 2)Local Correctness: Establishes annotation validity for each thread.
 3)Stability:         Ensures that each (local)thread annotation is interference-free under the execution of other threads.
*)

(*Store buffering litmus test*)
(*Program outline
  x := 1       |   y := 1 
  mfence       |   mfence
  r1 \<leftarrow> y      |   r2 \<leftarrow> x
 *)
 


theory Store_Buffering
imports   "../pierogi/Proof_Rules" 
begin

datatype PC =
  L1
| L2
| L3
| L4


consts t1 ::  TId 
consts t2 ::  TId
consts x :: Loc
consts y :: Loc
consts z :: Loc
consts w :: Loc
consts r1 :: Reg
consts r2 :: Reg 
consts r3 :: Reg 
consts r4 :: Reg 

definition "thrdsvars \<equiv> t1 \<noteq> t2 \<and> 
                         x \<noteq> y  \<and>
                         z \<noteq> y \<and>
                         z \<noteq> x  \<and>
                         w \<noteq> x \<and>
                         w \<noteq> y \<and>
                         w \<noteq> z \<and>
                         r1 \<noteq> r2 
                        " 

(*the state of the program*)
record mp_state =
  pc :: "TId \<Rightarrow> PC"
  state :: TState
  aux1 :: nat 
  aux2 :: nat 
  aux3 :: nat 
  aux4 :: nat

(*auxiliary definition for updating the pc*)
definition 
  "update_pc t nv pcf \<equiv> pcf (t := nv)"

lemmas mp_sis [simp] = 
  update_pc_def 
  thrdsvars_def

(*program encoding*)
definition prog :: "TId \<Rightarrow>  mp_state \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog t ts ts' \<equiv> 
  if t = t1
   then
    (case (pc ts) t of 
      L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [x:=1]\<^sub>t (state ts') \<and>
      (aux1 ts') = ( aux2 ts)  + 1  \<and>
      ( aux2 ts')  =  ( aux2 ts) \<and>
       aux3 ts'  =  aux3 ts  \<and>
       aux4 ts = aux4  ts'

    | L2 \<Rightarrow> pc ts' = update_pc t L3 (pc ts) \<and>
     (state ts)  [mfence]\<^sub>t  (state ts') \<and>
     aux1 ts = aux1 ts' \<and>
     aux2 ts = aux2  ts' \<and>
    aux3 ts'  =  aux3 ts  \<and>
     aux4 ts = aux4  ts'

   | L3 \<Rightarrow>  pc ts' = update_pc t L4 (pc ts) \<and>
        (state ts) [r1   \<leftarrow> y ]\<^sub>t (state ts')  \<and>
       aux1 ts = aux1 ts' \<and>
       aux2 ts = aux2  ts'\<and>
       aux3 ts'  =  1  \<and>
       aux4 ts = aux4  ts'

    | _ \<Rightarrow> ts = ts')
  else 
    if t = t2
    then 
  (case (pc ts) t of 

     L1 \<Rightarrow> pc ts'  = update_pc t  L2 (pc ts) \<and>
      (state ts) [y:=1]\<^sub>t (state ts') \<and>
      (aux2 ts') = ( aux1 ts)  + 1  \<and>
      ( aux1 ts')  =  ( aux1 ts) \<and>
      aux3 ts'  =  aux3 ts  \<and>
     aux4 ts = aux4  ts'
 
     | L2 \<Rightarrow>  pc ts' = update_pc t L3 (pc ts) \<and>
               (state ts)  [mfence]\<^sub>t  (state ts') \<and>
               aux1 ts = aux1 ts' \<and>
               aux2 ts = aux2  ts' \<and>
               aux3 ts'  =  aux3 ts  \<and>
               aux4 ts = aux4  ts'
 
     | L3 \<Rightarrow> pc ts' = update_pc t L4 (pc ts) \<and>
        (state ts) [r2   \<leftarrow> x ]\<^sub>t (state ts')  \<and>
       aux1 ts = aux1 ts' \<and>
       aux2 ts = aux2  ts'\<and>
       aux4 ts'  =  1  \<and>
       aux3 ts = aux3  ts'
    |_ \<Rightarrow> ts = ts'

  )
    else ts = ts'"


(*program invariant*)
definition prog_inv :: "TId \<Rightarrow> PC \<Rightarrow> mp_state \<Rightarrow> bool " where
"prog_inv t p ts \<equiv>
 if t = t1
   then 

   (case p of
          L1 \<Rightarrow> 
              ( aux1 ts = 0 \<and> aux2 ts = 0 \<and> aux4 ts = 0  )
              \<or> ( aux1 ts = 0 \<and> aux2 ts = 1  \<and> \<lceil>y,1\<rceil>\<^sub>t2 (state ts) )
 
                
        | L2 \<Rightarrow>
                ( aux1 ts = 2 \<and> aux2 ts = 1 \<and> \<lceil>y,1\<rceil>\<^sub>t2 (state ts)   \<and> \<lceil>MFENCE y\<rceil>\<^sub>t (state ts) ) \<or>
                 ( aux1 ts = 1 \<and> (aux2 ts = 0 \<or> aux2 ts = 2) \<and> ( (regs (state ts) t2  r2 = 1)  \<or> aux4 ts = 0) )

        | L3 \<Rightarrow>  
                ( aux1 ts = 2 \<and> aux2 ts = 1\<and> [y]\<^sub>t (state ts) =  {1}  ) \<or>
                 ( aux1 ts = 1 \<and>(aux2 ts = 0 \<or> aux2 ts = 2) \<and>(  (regs (state ts) t2  r2 = 1)  \<or> aux4 ts = 0) )

        | L4 \<Rightarrow>  
                ( aux1 ts = 2 \<and> aux2 ts = 1\<and>  (regs (state ts) t  r1 = 1)   ) \<or>
                  ( aux1 ts = 1 \<and> (aux2 ts = 0 \<or> aux2 ts = 2) \<and>(  (regs (state ts) t2  r2 = 1) \<or>  aux4 ts = 0)) 



      )
  else
 if t = t2 then
      (case p of
          L1 \<Rightarrow> 
                 ( aux2 ts = 0 \<and> aux1 ts = 0 \<and>  aux3 ts = 0  ) 
              \<or> ( aux2 ts = 0 \<and> aux1 ts = 1 \<and> \<lceil>x,1\<rceil>\<^sub>t1 (state ts))

        | L2 \<Rightarrow>  
                ( aux2 ts = 2 \<and> aux1 ts = 1 \<and> \<lceil>x,1\<rceil>\<^sub>t1 (state ts) \<and>  \<lceil>MFENCE x\<rceil>\<^sub>t (state ts) ) \<or>
                 ( aux2 ts = 1 \<and>  (aux1 ts = 2 \<or> aux1 ts = 0)\<and> (  (regs (state ts) t1  r1 = 1)   \<or>  aux3 ts = 0) )

        | L3 \<Rightarrow>   
                ( aux2 ts = 2 \<and> aux1 ts = 1\<and> [x]\<^sub>t (state ts) =  {1} ) \<or>
                 ( aux2 ts = 1 \<and>  (aux1 ts = 2 \<or> aux1 ts = 0) \<and> (  (regs (state ts) t1  r1 = 1)   \<or>  aux3 ts = 0) )

        | L4 \<Rightarrow> 
                ( aux2 ts = 2 \<and> aux1 ts = 1\<and>  (regs (state ts) t  r2 = 1)  ) \<or>
                 ( aux2 ts = 1 \<and> (aux1 ts = 2 \<or> aux1 ts = 0) \<and> (  (regs (state ts) t1  r1 = 1)   \<or>  aux3 ts = 0)) 

      )       
  else False " 


(*Initial State*)
definition init  :: "mp_state  \<Rightarrow> bool"   
  where
    "init s \<equiv> (pc s) t1 = L1 
      \<and> (pc s) t2 = L1
      \<and> aux1 s  = 0
      \<and> aux2 s = 0
      \<and> aux3 s = 0
      \<and> aux4 s = 0
      \<and> start (state s)"


lemma reg_same_st_low:
 assumes "total_wfs ts"
 and "step t ( Store  x v) ts ts'"
shows "  regs ts t' r   = regs ts' t' r"
  using assms(1) assms(2) reg_same_st total_wfs_def by auto

lemma reg_same_fl_low:
 assumes "total_wfs ts"
 and "step t ( Flush  x ) ts ts'"
shows "  regs ts t' r   = regs ts' t' r"
  using assms(1) assms(2) reg_same_fl total_wfs_def by auto

lemma reg_same_mf_low:
 assumes "total_wfs ts"
 and "step t ( mfence ) ts ts'"
shows "  regs ts t' r   = regs ts' t' r"
  using assms(2) reg_same_mfence by auto

(*Local Correctness of thread 1*)
lemma t1_local:
  assumes "thrdsvars"
      and "total_wfs  (state s)"
      and "prog_inv t1 ((pc s) t1) s" 
      and "prog t1 s s'"
    shows "prog_inv t1 ((pc s') t1) s'"
  using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t1", simp_all)
    apply (metis st_m_ord_lc st_ots_lm_ni st_ov_ni)
   apply (elim conjE disjE)
  apply (meson assms(2) mf_ov_ord total_wfs_def)
  apply (metis assms(2) reg_same_mf_low)
   apply blast
    apply (metis assms(2) reg_same_mf_low)
   apply linarith
  by (metis (no_types, lifting) ld_ov_lc reg_same_ld_dt singletonD)
  

(*Local Correctness of thread 2*)
lemma t2_local:
  assumes "thrdsvars"
      and "total_wfs (state s)"
      and "prog_inv t2 ((pc s) t2) s" 
      and "prog t2 s s'"
    shows "prog_inv t2 ((pc s') t2) s'"
  using assms 
   using assms 
  apply (simp add: prog_def prog_inv_def total_wfs_def)
  apply (cases "pc s t2", simp_all)
    apply (metis st_m_ord_lc st_ots_lm_ni st_ov_ni)
   apply (elim conjE disjE)
  apply (meson assms(2) mf_ov_ord total_wfs_def)
  apply (metis assms(2) reg_same_mf_low)
   apply blast
    apply (metis assms(2) reg_same_mf_low)
   apply linarith
  by (metis (no_types, lifting) ld_ov_lc reg_same_ld_dt singletonD)
  

(*Stability check for thread 1*)
lemma t1_global :
  assumes  "total_wfs  (state s)"
  and  "prog_inv t1 ((pc s) t1) s"
  and "prog_inv t2 ((pc s) t2) s"
  and "thrdsvars"
  and "prog t2 s s'"
shows  "prog_inv t1 ((pc s') t1) s'"
  using assms
 apply (simp add: prog_inv_def prog_def total_wfs_def  )
  apply (cases "pc s t1"; cases "pc s t2"; simp)
             apply (metis st_ots_lm_lc st_ov_lc vbounded_preserved)
            apply (meson assms(1) mf_ots_lm_s mf_ots_s total_wfs_def)
  apply (metis assms(1) ld_ots_lm_s ld_ots_s le_numeral_extra(4) not_one_le_zero numeral_1_eq_Suc_0 numerals(1) total_wfs_def)
  apply (metis Zero_neq_Suc reg_same_st)    apply (elim disjE)
            apply linarith
           apply (smt (verit, ccfv_SIG) assms(1) mf_ots_lm_ord_ni mf_ots_lm_s mf_ots_s total_wfs_def)
          apply (metis assms(1) reg_same_mf_low)
         apply (metis assms(1) reg_same_mf_low)
        apply (elim disjE)
           apply (metis n_not_Suc_n numeral_2_eq_2) 
  apply (metis (no_types, lifting) assms(1) ld_m_ord_ni ld_ots_lm_s ld_ots_s total_wfs_def)
         apply (metis ld_ov_lc singleton_iff)
        apply (metis n_not_Suc_n numeral_2_eq_2) 
  apply (metis nat.distinct(1) reg_same_st)
      apply (elim disjE)
         apply linarith
        apply (metis assms(1) mf_ots_s total_wfs_def)
       apply (metis assms(1) reg_same_mf_low)
      apply (metis assms(1) reg_same_mf_low)
  apply (metis ld_ov_lc ld_ov_ni numeral_1_eq_Suc_0 numeral_eq_iff semiring_norm(85) singleton_iff zero_neq_numeral)
    apply (metis (mono_tags, lifting) reg_same_st)
   apply (metis assms(1) reg_same_mf_low)
  by (smt (z3) empty_iff insert_iff ld_ov_lc n_not_Suc_n numeral_2_eq_2 reg_same_ld_dt)


(*Stability check for thread 2*)
lemma t2_global :
  assumes  "total_wfs  (state s)"
  and "prog_inv t2 ((pc s) t2) s"
  and "prog_inv t1 ((pc s) t1) s"
  and "thrdsvars"
  and "prog t1 s s'"
shows  "prog_inv t2 ((pc s') t2) s'" 
  using assms 
  apply (simp add: prog_inv_def prog_def total_wfs_def  )
 apply (cases "pc s t2"; cases "pc s t1"; simp)
     apply (metis st_ots_lm_lc st_ov_lc vbounded_preserved)
            apply (meson assms(1) mf_ots_lm_s mf_ots_s total_wfs_def)
  apply (metis assms(1) ld_ots_lm_s ld_ots_s le_numeral_extra(4) not_one_le_zero numeral_1_eq_Suc_0 numerals(1) total_wfs_def)
          apply (metis Zero_not_Suc reg_same_st)   
         apply (elim disjE)
            apply linarith
           apply (smt (verit, ccfv_SIG) assms(1) mf_ots_lm_ord_ni mf_ots_lm_s mf_ots_s total_wfs_def)
          apply (metis assms(1) reg_same_mf_low)
         apply (metis assms(1) reg_same_mf_low)
        apply (elim disjE)
           apply (metis n_not_Suc_n numeral_2_eq_2) 
  apply (metis (no_types, lifting) assms(1) ld_m_ord_ni ld_ots_lm_s ld_ots_s total_wfs_def)
         apply (metis ld_ov_lc singleton_iff)
        apply (metis n_not_Suc_n numeral_2_eq_2) 
  apply (metis Zero_not_Suc reg_same_st)
      apply (elim disjE)
         apply linarith
        apply (metis assms(1) mf_ots_s total_wfs_def)
       apply (metis assms(1) reg_same_mf_low)
      apply (metis assms(1) reg_same_mf_low)
  apply (metis ld_ov_lc ld_ov_ni numeral_1_eq_Suc_0 numeral_eq_iff semiring_norm(85) singleton_iff zero_neq_numeral)
    apply (metis (mono_tags, lifting) reg_same_st)
   apply (metis assms(1) reg_same_mf_low)
  by (smt (z3) empty_iff insert_iff ld_ov_lc n_not_Suc_n numeral_2_eq_2 reg_same_ld_dt)


(*Initialization*)
theorem init_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "init s"
shows "prog_inv t ((pc s) t) s"
  using assms 
    apply (simp add: init_def  prog_inv_def )  
  by blast


theorem final_inv:
  assumes "total_wfs (state s)"   
  and "thrdsvars"
  and "t \<in> {t1, t2}"
  and "t' \<in> {t1, t2}"
  and "prog_inv t ((pc s) t) s"
  and "prog_inv t' ((pc s) t') s" 
  and "prog t' s s'"
shows "prog_inv t ((pc s') t) s'"
  using assms apply (simp del: thrdsvars_def) 
  using t1_global t1_local t2_global t2_local
  by blast
 

end




