(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

(*Paper\Isabelle correspondence
  Paper                                                         Isabelle

  [TS]\<^sub>\<tau>\<^sup>O\<^sup>F(\<sigma>, x, vrnew) or [TS]\<^sub>\<tau>(\<sigma>, x) :  the set of timestamps that are observable for \<tau> to read for location x in state \<sigma>
                                                                | OTS \<sigma> x: OTS set (defined in Language.thy)

  [x]\<^sub>\<tau> \<sigma>: coherence\thread  view of thread \<tau> and  address x     |[x]\<^sub>\<tau> \<sigma>: OV set 

  [TS]\<^sub>\<tau>\<^sup>A(\<sigma>, x): the timestamps of the asynchronous view of \<tau> in location x and state \<sigma>.
                                                                | OATS  \<sigma> \<tau> x: OATS set
 
  [x]\<^sup>A\<^sub>\<tau> \<sigma>: asynchronous memory view of thread \<tau> and address x    |[x]\<^sup>A\<^sub>\<tau> \<sigma>: OAV set

  [TS]\<^sup>P(\<sigma>, x): the timestamps of the persistent view in location x and state \<sigma>.
                                                                | OPTS  \<sigma> x: OPTS set

  [x]\<^sup>P \<sigma>: persistent memory view of address x                    |[x]\<^sub>P \<sigma>: OPV set  

  [TS]\<^sup>o\<^sup>v (\<sigma>, x, v): the set of timestamps that \<tau> can observe for x with value v
                                                                 | cond_ts \<sigma> \<tau> x v: cond_ts set

  [TS]\<^sup>C\<^sup>O(\<sigma>, x, v, y): the set of timestamps that \<tau> can observe for y, from any timestamp t \<in>  [TS]\<^sup>o\<^sup>v (\<sigma>, x, v).
                                                                 | COBTS x y v \<tau> \<sigma>: COBTS set

  \<langle>x  v\<rangle>[y]\<^sub>\<tau>: returns a set of values that \<tau> may read for y after it reads value v for x    
                                                                 |\<langle>x  v\<rangle>[y]\<^sub>\<tau>:  COBV set

  \<lbrakk>x\<rbrakk>\<^sub>\<tau> \<sigma>: holds iff \<tau> is currently viewing the  last  write to x  |\<lceil>x\<rceil>\<^sub>\<tau> \<sigma>

  \<lbrakk>x\<rbrakk>\<^sup>F\<^sub>\<tau> \<sigma>: holds iff a flush of x by \<tau>  is  guaranteed to flush the last write to x to persistent
  memory                                                         |\<lceil>FLUSH x\<rceil>\<^sub>\<tau> \<sigma>

  \<lbrakk>x\<rbrakk>\<^sup>M\<^sub>\<tau> \<sigma>: holds iff after performing an mfence operation, the view of thread Τ will be the last write on x.
                                                                 |\<lceil>MFENCE x\<rceil>\<^sub>\<tau> \<sigma>

  |x,v| \<sigma>: returns the number of writes to x with value v        |\<lparr>x,v\<rparr> \<sigma>
  
  \<lbrakk>x:v\<rbrakk> \<sigma>: holds if the last written value on x is v              |\<lceil>x:v\<rceil> \<sigma> (defined in Language.thy) *)


theory Assertions
imports  Language Main
begin

(*Set of observable persistent timestamps: OPTS*)
definition "OPTS  ts  l  \<equiv> {t | t.  0\<le> t \<and> t < length(memory ts) \<and> l= comploc ((memory ts)!t) l
  \<and> ( notoverwritten ts (maxvp ts  l)  t l)} "


(*Set of observable asynchronous timestamps: OATS*)
definition "OATS  ts ti  l  \<equiv> {t | t.  0\<le> t \<and> t < length(memory ts) \<and> l= comploc ((memory ts)!t) l
   \<and> ( notoverwritten ts (vpasync ts ti  l)  t l)} " 


(*Auxiliary definition for the set of conditional observation*)
definition " compvrnew ts ti ind l \<equiv> 
 let  nview =  if( ind \<noteq> (coh ts ti) l )
                   then ind 
                else vrnew ts ti
in  max nview (vrnew ts ti)"


(*Auxiliary definition for the set of conditional observation*)
definition " cond_ts ts ti l v \<equiv> {compvrnew ts ti t l   | t. t \<in> OTS ts ti l \<and> v = compval  ((memory ts)!t)}"


(*Set of conditionally observable timestamps: COBTS*)
definition " COBTS x y v ti ts \<equiv>  \<Union> { OTSF ts ti y nview | nview.   nview  \<in> cond_ts ts ti x v} "


(*Auxiliary definition for sets of values. Maps an index in the memory array to the value of the message that this index points to *)
definition mapval
  where "mapval  B li = (\<lambda>x. compval( (li)!x) ) ` B"  

(*Abbreviation for the set of observable asynchronous values: OAV*)
abbreviation OAab ::" Loc \<Rightarrow> TId \<Rightarrow> TState \<Rightarrow> Val set" ("[_]\<^sup>A\<^sub>_  _" [100,100,100])
where
"[l]\<^sup>A\<^sub>t \<sigma> \<equiv> mapval (OATS  \<sigma> t  l) (memory \<sigma>) "

(*Abbreviation for the set of observable persistent values: OPV*)
abbreviation OPab:: "Loc \<Rightarrow> TState \<Rightarrow> Val set " ("[_]\<^sub>P_"[100,100])
where
"[l]\<^sub>P \<sigma> \<equiv> mapval (OPTS  \<sigma>  l) (memory \<sigma>) "

(*Abbreviation for the set of observable values: OV*)
abbreviation OVab ::" Loc \<Rightarrow> TId \<Rightarrow> TState \<Rightarrow> Val set" ("[_]\<^sub>_ _" [100,100,100])
  where
 "[l]\<^sub>t \<sigma> \<equiv> mapval (OTS  \<sigma> t l) (memory \<sigma>)  "


(*Abbreviation for the  set of values of the conditional observation: COBV*)
abbreviation COBVab :: "Loc  \<Rightarrow> Val  \<Rightarrow> Loc  \<Rightarrow> TId \<Rightarrow> TState \<Rightarrow> Val set" ("\<langle>_  _\<rangle>[_]\<^sub>_ _ " [100, 100, 100, 100, 100])
  where "\<langle>x  v\<rangle>[y]\<^sub>t  \<sigma> \<equiv>  mapval (COBTS x y v t \<sigma>) (memory \<sigma>) "


(*Abbreviation for set of observable timestamps containing only the last write at x*)
abbreviation OTS_local_max ::" Loc \<Rightarrow> TId \<Rightarrow> TState \<Rightarrow> bool" ("\<lceil>_\<rceil>\<^sub>_ _" [100,100,100])
  where
 "\<lceil>l\<rceil>\<^sub>t \<sigma> \<equiv>  (OTS  \<sigma> t l) = { last_entry  \<sigma>  l  }  "


(*Combining  OTS_local_max and  OV in one abbreviation to save space*)
abbreviation OV_local_max_comp ::" Loc \<Rightarrow>Val \<Rightarrow> TId \<Rightarrow> TState \<Rightarrow> bool" ("\<lceil>_,_\<rceil>\<^sub>_ _" [100, 100,100,100])
  where
 "\<lceil>x,v\<rceil>\<^sub>t \<sigma> \<equiv> [x]\<^sub>t \<sigma> ={v} \<and>  \<lceil>x\<rceil>\<^sub>t \<sigma> "


(*Expression that holds iff a flush on address x is  guaranteed to flush the last  write to x to persistent memory.*)
abbreviation flush_order :: " Loc \<Rightarrow> TId \<Rightarrow>  TState \<Rightarrow> bool" ( "\<lceil>FLUSH _ \<rceil>\<^sub>_ _ "  [100,100,100])
 where
 "\<lceil>FLUSH l\<rceil>\<^sub>t \<sigma> \<equiv>   last_entry  \<sigma>  l \<le>   (max ( maxcoh \<sigma> t)  (maxvp \<sigma> l))   "


(*Expression that holds iff after performing an mfence, the view of thread t will be the last write on address l*)
abbreviation mfence_order :: " Loc \<Rightarrow> TId \<Rightarrow>  TState \<Rightarrow> bool" ( "\<lceil>MFENCE _ \<rceil>\<^sub>_ _ "  [100,100,100])
 where
 "\<lceil>MFENCE l\<rceil>\<^sub>t \<sigma> \<equiv>   last_entry  \<sigma>  l \<le>  (maxcoh \<sigma> t) " 


(*Auxiliary definition for count_oc*)
definition " oc_set ts  x v \<equiv>  {t | t.  0\<le> t \<and> t < length(memory ts) \<and> x= comploc ((memory ts)!t) x \<and> compval  ((memory ts)!t) = v}"

(*Returns the number of memory messages with location x and value v *)
abbreviation count_oc :: " Loc \<Rightarrow> Val \<Rightarrow> TState \<Rightarrow> nat " ( "\<lparr>_,_\<rparr>  _" [100,100,100])
  where
" \<lparr>x,v\<rparr>  \<sigma>  \<equiv> card  (oc_set \<sigma>  x v) " 


(*auxiliary lemmas*)
lemma st_no:
  assumes "t = 0"
  and " start ts"
shows "\<forall>l ti.  notoverwritten ts (vrnew ts ti) t l"
  using assms  apply (simp add: start_def notoverwritten_def)
  by blast

lemma [simp]: "OTS (flush_trans ti l ts) ti l = OTS ts ti l"
  apply (simp add: flush_trans_def OTS_def OTSF_def notoverwritten_def  Let_def)
  done  



end